"""
Contains definitions for the CircuitPython APIs used on Adafruit boards so they
can be used in the editor for autocomplete and call tips.

Copyright (c) 2015-2017 Nicholas H.Tollervey and others (see the AUTHORS file).

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
Documentation pulled from: https://docs.micropython.org/en/latest/pyboard/
"""

PYBOARD_APIS = [
    _(
        "pyb.delay(ms) \nDelay for the given number of milliseconds.\n\nParameters:\nms (int) -- the number of milliseconds to delay.\n"
    ),
    _(
        "pyb.udelay(us) \nDelay for the given number of microseconds.\n\nParameters:\nus (int) -- the number of microseconds to delay.\n"
    ),
    _(
        "pyb.millis() \nReturns the number of milliseconds since the board was last reset.\n\nReturns: milliseconds since reset.\n\nReturn type: int\n"
    ),
    _(
        "pyb.micros() \nReturns the number of microseconds since the board was last reset.\n\nReturns: microseconds since reset.\n\nReturn type: int\n"
    ),
    _(
        "pyb.elapsed_millis(start) \nReturns the number of milliseconds which have elapsed since start.\n\nParameters:\nstart (int) -- the number of milliseconds to start counting from.\n\nReturns: milliseconds since start.\n\nReturn type: int\n"
    ),
    _(
        "pyb.elapsed_micros(start) \nReturns the number of microseconds which have elapsed since start.\n\nParameters:\nstart (int) -- the number of microseconds to start counting from.\n\nReturns: microseconds since start.\n\nReturn type: int\n"
    ),
    _(
        "pyb.hard_reset() \nResets the board in a manner similar to pushing the external RESET button.\n"
    ),
    _("pyb.bootloader() \nActivate the bootloader without BOOT* pins.\n"),
    _(
        "pyb.fault_debug(valus) \nEnable or disable hard-fault debugging.\nA hard-fault is when there is a fatal error in the underlying system, like an\ninvalid memory access.\n\nParameters:\nvalue (bool) -- \n\nFalse: board will automatically reset if there is a hard fault\nTrue: when the board has a hard fault, it will print the registers and the stack\n\ttrace, and then cycle the LEDs indefinitely\n"
    ),
    _(
        "pyb.disable_irq() \nDisable interrupt requests.\n\nReturns: Previous IRQ state.\n\nReturn type: bool\n"
    ),
    _(
        "pyb.enable_irq(state=True) \nEnable interrupt requests.\n\nParameters:\nstate (bool) -- Whether the IRQs are enabled\n"
    ),
    _(
        "pyb.freq([sysclk[, hclk[, pclk1[, pclk2]]]]) \nIf given no arguments, returns a tuple of clock frequencies:\n(sysclk, hclk, pclk1, pclk2). Otherwise, sets the frequency of the CPU, and the\nbusses if additional arguments are given.\n\nParameters:\nsysclk (int) [optional] -- frequency of the CPU in MHz.\nhclk (int) [optional] -- frequency of the AHB bus, core memory and DMA.\npclk1 (int) [optional] -- frequency of the APB1 bus.\npclk2 (int) [optional] -- frequency of the APB2 bus.\n\nReturns: clock frequencies.\n\nReturn type: tuple\n"
    ),
    _(
        "pyb.wfi() \nWait for an internal or external interrupt.\nThis function will block for at most 1ms.\n"
    ),
    _(
        "pyb.stop() \nPut the board in a “sleeping” state. To wake from this sleep state requires an\nexternal interrupt or a real-time-clock event. Upon waking execution continues\nwhere it left off.\n"
    ),
    _(
        "pyb.standby() \nPut the board into a “deep sleep” state. To wake from this sleep state requires\na real-time-clock event, or an external interrupt on X1 (PA0=WKUP) or\nX18 (PC13=TAMP1). Upon waking the system undergoes a hard reset.\n"
    ),
    _(
        "pyb.info([dump_alloc_table]) \nPrint out lots of information about the board.\n\nParameter:\ndump_alloc_table (bool) [optional] -- Whether to dump the alloc table.\n"
    ),
    _(
        "pyb.main(filename) \nSet the filename of the main script to run after boot.py is finished.\n\nIt only makes sense to call this function from within boot.py.\n\nParameter:\nfilename (string) -- name of the desired main file.\n"
    ),
    _(
        "pyb.mount(device, mountpoint, *, readonly=False, mkfs=False) \nMount a block device and make it available as part of the filesystem.\n\nSee online documentation for full instructions.\n"
    ),
    _(
        "pyb.repl_uart(uart) \nGet or set the UART object where the REPL is repeated on.\n\nParameter:\nuart (pyb.UART) -- The uart to use\n."
    ),
    _(
        "pyb.rng() \nReturn a 30-bit hardware generated random number.\n\nReturns: 30-bit random number.\n\nReturn value: int\n"
    ),
    _("pyb.sync() \nSync all file systems.\n"),
    _(
        "pyb.unique_id() \nReturns a string of 12 bytes (96 bits), which is the unique ID of the MCU.\n\nReturns: unique ID of the MCU.\n\nReturn value: string\n"
    ),
    _(
        "pyb.usb_mode([modestr, ]vid=0xf055, pid=0x9801, hid=pyb.hid_mouse) \nIf called with no arguments, return the current USB mode as a string. If called\nwith modestr provided, attempts to set USB mode.\n\nThis can only be done when called from boot.py before pyb.main() has been called.\n\nSee online documentation for full instructions.\n"
    ),
    _("pyb.Accel() \nCreate an accelerometer object.\n"),
    _(
        "pyb.Accel.filtered_xyz() \nGet filtered x, y and z values.\n\nReturns: accelerometer x, y, and z values.\n\nReturn type: 3-tuple\n"
    ),
    _(
        "pyb.Accel.tilt() \nGet the tilt register.\n\nReturns: contents of tilt register,\n\nReturn type: int\n"
    ),
    _(
        "pyb.Accel.x() \nGet the x-axis register.\n\nReturns: contents of x-axis register,\n\nReturn type: int\n"
    ),
    _(
        "pyb.Accel.y() \nGet the y-axis register.\n\nReturns: contents of y-axis register,\n\nReturn type: int\n"
    ),
    _(
        "pyb.Accel.z() \nGet the z-axis register.\n\nReturns: contents of z-axis register,\n\nReturn type: int\n"
    ),
    _(
        "pyb.ADC(pin) \nAn Analog to Digital conversion object associated with the given pin\n\nParameter:\npin (pyb.Pin) -- the pin to read values from\n"
    ),
    _(
        "pyb.ADC.read() \nRead the value on the analog pin and return it.\nThe returned value will be between 0 and 4095.\n\nReturns: analog value of pin.\n\nReturn type: int\n"
    ),
    _(
        "pyb.ADC.read_timed(buf, timer) \nRead analog values into buf at a rate set by the timer object.\n\nParameters:\nbuf (bytearray or array.array) -- The ADC values have 12-bit resolution and are\nstored directly into buf if its element size is 16 bits or greater. If buf has\nonly 8-bit elements (eg a bytearray) then the sample resolution will be reduced\nto 8 bits.\ntimer (pyb.Timer) -- a Timer object, a sample is read each time the timer triggers."
    ),
    _(
        "pyb.CAN(bus, ...) \nConstruct a CAN object on the given bus. With no additional parameters, the CAN\nobject is created but not initialised. If extra arguments are given, the bus is\ninitialised. See CAN.init() for parameters of initialisation.\n\nParameters:\nbus can be 1-2, or 'YA' or 'YB'."
    ),
    _(
        "pyb.CAN.init(mode, extframe=False, prescaler=100, *, sjw=1, bs1=6, bs2=8, auto_restart=False) \nInitialise the CAN bus with the given parameters.\n\nParameters:\nmode (int) -- one of: NORMAL, LOOPBACK, SILENT, SILENT_LOOPBACK. These are\nincluded in the pyb.CAN class as class variables.\nextframe (bool) -- if extframe is True then the bus uses extended identifiers in\nthe frames (29 bits); otherwise it uses standard 11 bit identifiers.\nprescaler (int) -- used to set the duration of 1 time quanta; the time quanta\nwill be the input clock divided by the prescaler.\nsjw (int) -- is the resynchronisation jump width in units of the time quanta; it\ncan be 1, 2, 3, 4\nbs1 (int) -- defines the location of the sample point in units of the time\nquanta; it can be between 1 and 1024 inclusive.\nbs2 (int) -- defines the location of the transmit point in units of the time\nquanta; it can be between 1 and 16 inclusive.\nauto_restart (bool) -- sets whether the controller will automatically try and\nrestart communications after entering the bus-off state.\n"
    ),
    _("pyb.CAN.deinit() \nTurn off the CAN bus.\n"),
    _(
        "pyb.CAN.restart() \nForce a software restart of the CAN controller without resetting its configuration.\n"
    ),
    _(
        "pyb.CAN.state() \nReturn the state of the controller.\n\nReturns: the state of the controller.\nCan be one of the following: CAN.STOPPED, CAN.ERROR_ACTIVE, CAN.ERROR_WARNING,\nCAN.ERROR_PASSIVE, CAN.BUS_OFF.\n\nReturn type: int\n"
    ),
    _(
        "pyb.CAN.info([list]) \nGet information about the controller’s error states and TX and RX buffers.\n\nParameters:\nlist (list) [optional] -- If list is provided then it should be a list object\nwith at least 8 entries, which will be filled in with the information.\n\nReturns: If list parameter is not provided, a new list will be created and filled in.\n\nReturn type: list\n"
    ),
    _(
        "pyb.CAN.setfilter(bank, mode, fifo, params, *, rtr) \nConfigure a filter bank\n\nParameters:\nbank (int) -- the filter bank that is to be configured\nmode (int) -- the mode the filter should operate in. One of the following:\nCAN.LIST16, CAN.LIST32, CAN.MASK16, CAN.MASK32\nfifo (int) -- which fifo (0 or 1) a message should be stored in, if it is\naccepted by this filter\nparams (array (int)) -- values that defines the filter. The contents of the array\ndepends on the mode argument\nrtr (array (bool)) [optional]  -- array that states if a filter should accept a\nremote transmission request message. If this argument is not given then it defaults to False for all entries."
    ),
    _(
        "pyb.CAN.clearfilter(bank) \nClear and disables a filter bank\n\nParameters:\nbank (int) -- the filter bank that is to be cleared.\n"
    ),
    _(
        "pyb.CAN.any(fifo) \nWhether any message waiting on the FIFO\n\nParameters:\nfifo (int) -- the FIFO to check.\n\nReturns: Whether a message is waiting\n\nReturn type: bool\n"
    ),
    _(
        "pyb.DAC(port, bits=8, *, buffering=None) \nConstruct a new DAC object on the given port.\n\nParameters:\nport (pyb.Pin or int) -- can be a pin object, or an integer (1 or 2). DAC(1) is\non pin X5 and DAC(2) is on pin X6.\nbits (int) -- The resolution, and can be 8 or 12. The maximum value for the write\nand write_timed methods will be 2**bits-1\nbuffering (bool or None) -- The buffering parameter selects the behaviour of the\nDAC op-amp output buffer, whose purpose is to reduce the output impedance. It can\nbe None to select the default, False to disable buffering completely, or True to\nenable output buffering.\n"
    ),
    _(
        "pyb.DAC.init(bits=8, *, buffering=None) \nReinitialise the DAC.\n\nParameters:\nbits (int) -- The resolution, and can be 8 or 12. The maximum value for the write and write_timed methods will be 2**bits-1\nbuffering (bool or None) -- The buffering parameter selects the behaviour of the DAC op-amp output buffer, whose purpose is to reduce the output impedance. It can be None to select the default, False to disable buffering completely, or True to enable output buffering.\n"
    ),
    _(
        "pyb.DAC.deinit() \nDe-initialise the DAC making its pin available for other uses.\n"
    ),
    _(
        "pyb.DAC.noise(freq) \nGenerate a pseudo-random noise signal. A new random sample is written to the DAC\noutput at the given frequency.\n\nParameters:\nfreq (int) -- the frequency to generate noise at.\n"
    ),
    _(
        "pyb.DAC.triangle(freq) \nGenerate a triangle wave. The value on the DAC output changes at the given\nfrequency, and the frequency of the repeating triangle wave itself is 2048 times smaller.\n\nParameters: freq (int) -- the frequency to generate the wave at.\n"
    ),
    _(
        "pyb.DAC.write(value) \nDirect access to the DAC output. The minimum value is 0. The maximum value is 2**bits-1.\n\nParameters: value (int) -- the value to write to the DAC\n"
    ),
    _(
        "pyb.DAC.write_timed(data, freq, *, mode=DAC.NORMAL) \nInitiates a burst of RAM to DAC using a DMA transfer. The input data is treated\nas an array of bytes in 8-bit mode, and an array of unsigned half-words\n(array typecode ‘H’) in 12-bit mode.\n\nParameters:\ndata (array) -- data to be written to the DAC.\nfreq (int or Timer) -- frequency or Timer to determin how often to trigger DAC sample\nmode (constant) -- can be DAC.NORMAL or DAC.CIRCULAR."
    ),
    _(
        "pyb.ExtInt(pin, mode, pull, callback) \nCreate an ExtInt object.\n\nParameters:\npin (pyb.Pin) -- the pin on which to enable the interrupt.\nmode (constant) can be one of: ExtInt.IRQ_RISING - trigger on a rising edge;\nExtInt.IRQ_FALLING - trigger on a falling edge; ExtInt.IRQ_RISING_FALLING - trigger\non a rising or falling edge.\npull (constant) -- can be one of: - pyb.Pin.PULL_NONE - no pull up or down\nresistors; pyb.Pin.PULL_UP - enable the pull-up resistor;\npyb.Pin.PULL_DOWN - enable the pull-down resistor.\ncallback (function) -- the function to call when the interrupt triggers. The\ncallback function must accept exactly 1 argument, which is the line that triggered the interrupt.\n"
    ),
    _("pyb.ExtInt.regs() \nPrints the values of the EXTI registers.\n"),
    _(
        "pyb.ExtInt.disable() \nDisable the interrupt associated with the ExtInt object.\n"
    ),
    _("pyb.ExtInt.enable() \nEnable a disabled interrupt.\n"),
    _(
        "pyb.ExtInt.line() \nReturn the line number that the pin is mapped to.\n\nReturns: line number that the pin is mapped to\n\nReturn type: int\n"
    ),
    _("pyb.ExtInt.swint() \nTrigger the callback from software.\n"),
    _(
        "pyb.I2C(bus, ...) \nConstruct an I2C object on the given bus. bus can be 1 or 2, ‘X’ or ‘Y’. With no\nadditional parameters, the I2C object is created but not initialised. If extra\narguments are given, the bus is initialised. See init for parameters of initialisation.\n\nParameters:\nbus (int or string) -- the bus to attach to. Can be 1 or 2, ‘X’ or ‘Y’.\n"
    ),
    _("pyb.I2C.deinit() \nTurn off the I2C bus.\n"),
    _(
        "pyb.I2C.init(mode, *, addr=0x12, baudrate=400000, gencall=False, dma=False) \nInitialise the I2C bus with the given parameters\n\nParameters:\nmode (constant) -- must be either I2C.MASTER or I2C.SLAVE\naddr (int) -- the 7-bit address (only sensible for a slave)\nbaudrate (int) -- the SCL clock rate (only sensible for a master)\ngencall (bool) -- whether to support general call mode\ndma (bool) -- whether to allow the use of DMA for the I2C transfers\n"
    ),
    _(
        "pyb.I2C.is_ready(addr) \nCheck if an I2C device responds to the given address. Only valid when in master mode.\n\nParameters:\naddr (int) -- the address to check\n\nReturns: Whether the address responds\n\nReturn type: bool\n"
    ),
    _(
        "pyb.I2C.mem_read(data, addr, memaddr, *, timeout=5000, addr_size=8) \nRead from the memory of an I2C device\n\nParameters:\ndata (int or buffer) -- number of bytes to read or a buffer to read into\naddr (int) -- the I2C device address\nmemaddr (int) -- the memory location within the I2C device\ntimeout (int) the timeout in milliseconds to wait for the read\naddr_size (int) -- width of memaddr: 8 or 16 bits\n\nReturns: the read data. This is only valid in master mode.\n\nReturn type: bytes\n"
    ),
    _(
        "pyb.I2C.mem_write(data, addr, memaddr, *, timeout=5000, addr_size=8) \nWrite to the memory of an I2C device\nParameters:\ndata (int or buffer) -- number of bytes to write or a buffer to write into\naddr (int) -- the I2C device address\nmemaddr (int) -- the memory location within the I2C device\ntimeout (int) the timeout in milliseconds to wait for the write\naddr_size (int) -- width of memaddr: 8 or 16 bits\n"
    ),
    _(
        "pyb.I2C.recv(recv, addr=0x00, *, timeout=5000) \nReceive data on the bus.\n\nParameters:\nrecv (int or buffer) -- can be the number of bytes to receive, or a mutable\nbuffer, which will be filled with received bytes\naddr (int) -- the address to receive from (only required in master mode)\ntimeout (int) -- the timeout in milliseconds to wait for the receive\n\nReturns: if recv is an integer then a new buffer of the bytes received\n\nReturn type: bytes\n"
    ),
    _(
        "pyb.I2C.send(send, addr=0x00, *, timeout=5000) \nSend data on the bus.\n\nParameters:\nsend (int or buffer) -- the data to send\naddr (int) the address to send to (only required in master mode)\ntimeout (int) -- the timeout in milliseconds to wait for the send\n"
    ),
    _(
        "pyb.I2C.scan() \nScan all I2C addresses from 0x01 to 0x7f and return a list of those that respond.\nOnly valid when in master mode.\n\nReturns: valid I2C addresses on the bus\n\nReturn type: list\n"
    ),
    _(
        "pyb.LCD(skin_position) \nConstruct an LCD object in the given skin position.\n\nParameters:\nskin_position (string) -- can be ‘X’ or ‘Y’, and should match the position where\nthe LCD pyskin is plugged in.\n"
    ),
    _(
        "pyb.LCD.command(instr_data, buf) \nSend an arbitrary command to the LCD controller.\n\nParameters:\ninstr_data (int) -- 0 for instr_data to send an instruction, otherwise pass 1\nto send data\nbuf (int or buffer) -- a buffer with the instructions/data to send\n"
    ),
    _(
        "pyb.LCD.contrast(value) \nSet the contrast of the LCD\n\nParameters:\nvalue (int) -- the contrast value, valid values are between 0 and 47.\n"
    ),
    _(
        "pyb.LCD.fill(color) \nFill the screen with the given colour.\n\nParameters:\ncolor (int) -- 0 or 1 for white or black respectively.\n"
    ),
    _(
        "pyb.LCD.get(x, y) \nGet the pixel at the position (x, y)\n\nParameters:\nx (int) -- the X coordinate\ny (int) -- the Y coordinate\n\nReturns: the pixel value, either 0 or 1.\n\nReturn type: int\n"
    ),
    _(
        "pyb.LCD.light(value) \nTurn the backlight on/off.\n\nParameters:\nvalue (int or bool) -- True or 1 turns it on, False or 0 turns it off.\n"
    ),
    _(
        "pyb.LCD.pixel(x, y, colour) \nSet the pixel at (x, y) to the given colour (0 or 1).\n\nParameters:\nx (int) -- the X coordinate\ny (int) -- the Y coordinate\ncolor (int) -- the color.\n"
    ),
    _("pyb.LCD.show() \nShow the hidden buffer on the screen.\n"),
    _(
        "pyb.LCD.text(str, x, y, colour) \nDraw the given text to the position (x, y) using the given colour (0 or 1).\n\nParameters:\nstr (string) -- the text to display.\nx (int) -- the X coordinate\ny (int) -- the Y coordinate\ncolor (int) -- the color.\n"
    ),
    _(
        "pyb.LCD.write(str) \nWrite the string str to the screen. It will appear immediately.\n\nParameters:\nstr (string) -- the text to display.\n"
    ),
    _(
        "pyb.LED(id) \nCreate an LED object associated with the given LED\n\nParameters:\nid (int) -- the LED number, 1-4\n"
    ),
    _(
        "pyb.LED.intensity([value]) \nGet or set the LED intensity. If no argument is given, return the LED intensity.\n\nParameters:\nvalue (int) [optional] -- intensity value ranges between 0 (off) and 255 (full on)\n\nReturns: None or LED intensity\n\nReturn type: None or int\n"
    ),
    _("pyb.LED.off() \nTurn the LED off.\n"),
    _("pyb.LED.on() \nTurn the LED on.\n"),
    _(
        "pyb.LED.toggle() \nToggle the LED between on (maximum intensity) and off. If the LED is at non-zero\nintensity then it is considered “on” and toggle will turn it off.\n"
    ),
    _(
        "pyb.Pin(id, ...) \nCreate a new Pin object associated with the id. If additional arguments are given,\nthey are used to initialise the pin. See pyb.Pin.init()\n\nParameters:\nid (constant or string) -- the identifier of the pin to use\n"
    ),
    _(
        "pyb.Pin.debug([state]) \nGet or set the debugging state\n\nParameters:\nstate (bool) [optional] -- the debugging state to set, if any.\n\nReturns: whether the debugging state is set (if nothing is passed to the method)\n\nReturn type: bool\n"
    ),
    _(
        "pyb.Pin.init(mode, pull=Pin.PULL_NONE, af=-1) \nInitialise the pin\n\nParameters:\nmode (constant) -- can be one of: Pin.IN, Pin.OUT_PP, Pin.OUT_OD, Pin.AF_PP, Pin.AF_OD, Pin.ANALOG\npull (constant) -- can be one of: Pin.PULL_NONE, Pin.PULL_UP, Pin.PULL_DOWN\naf (int) -- when mode is Pin.AF_PP or Pin.AF_OD, then af can be the index or name\nof one of the alternate functions associated with a pin\n"
    ),
    _(
        "pyb.Pin.value([value]) \nGet or set the digital logic level of the pin\n\nParameters:\nvalue (int or bool) [optional] -- if value converts to True, the pin is set high,\notherwise it is set low\n\nReturns: with no argument, return 0 or 1 depending on the logic level of the pin\n\nReturn type: int\n"
    ),
    _(
        "pyb.Pin.af() \nReturns the currently configured alternate-function of the pin\n\nReturns: one of the integer representations of the allowed constants for af\n\nReturn type: int\n"
    ),
    _(
        "pyb.Pin.af_list() \nReturns an array of alternate functions available for this pin.\n\nReturns:array of alternate functions available\n\nReturn type: list\n"
    ),
    _(
        "pyb.Pin.gpio() \nReturns the base address of the GPIO block associated with this pin.\n\nReturns: the base address of the GPIO block\n\nReturn type: int"
    ),
    _(
        "pyb.Pin.mode() \nReturns the currently configured mode of the pin.\n\nReturns: integer returned will match one of the allowed constants for the mode\n\nReturn type: int"
    ),
    _(
        "pyb.Pin.name() \nReturns the pin name.\n\nReturns: the name of the pin\n\nReturn type: string"
    ),
    _(
        "pyb.Pin.names() \nReturns the board and cpu names for the pin.\n\nReturns: the names of the pin\n\nReturn type: list"
    ),
    _(
        "pyb.Pin.pin() \nReturns the CPU pin number.\n\nReturns: the number of the pin\n\nReturn type: int"
    ),
    _(
        "pyb.Pin.port() \nReturns the CPU port number.\n\nReturns: the number of the port\n\nReturn type: int"
    ),
    _(
        "pyb.Pin.pull() \nReturns the currently configured pull of the pin.\n\nReturns: The integer returned will match one of the allowed constants for the pull\n\nReturn type: int"
    ),
    _("pyb.RTC() \nCreate an RTC object.\n"),
    _(
        "pyb.RTC.datetime([datetimetuple]) \nGet or set the date and time of the RTC. With no arguments, this method returns\nan 8-tuple with the current date and time. With 1 argument (being an 8-tuple) it\nsets the date and time\n\nParameters:\ndatetimetuple (tuple) [optional] -- The 8-tuple has the following format:\n(year, month, day, weekday, hours, minutes, seconds, subseconds). All tuple\nelements are integers\n\nReturns: the current date and time\n\nReturn type: tuple (8 elements)\n"
    ),
    _(
        "pyb.RTC.wakeup(timeout, callback=None) \nSet the RTC wakeup timer to trigger repeatedly at every timeout milliseconds.\nThis trigger can wake the pyboard from both the sleep states: pyb.stop() and pyb.standby()\n\nParameters:\ntimeout (int or None) -- timeout in milliseconds or None to disable timer\ncallback (function or None) -- if callback is given then it is executed at every\ntrigger of the wakeup timer. callback must take exactly one argument\n"
    ),
    _(
        "pyb.RTC.info() \nGet information about the startup time and reset source\n\nReturns: RTC info: lower 0xffff are the number of milliseconds the RTC took to\nstart up, bit 0x10000 is set if a power-on reset occurred, bit 0x20000 is set if\nan external reset occurred\n\nReturn type: int\n"
    ),
    _(
        "pyb.RTC.calibration([cal]) \nGet or set RTC calibration.\n\nParameters: cal (int) [optional] -- The usable calibration range is: (-511 * 0.954) ~= -487.5 ppm up to (512 * 0.954) ~= 488.5 ppm\n\nReturns: with no arguments, the current calibration value, in the range [-511 : 512]\n\nReturn type: int\n"
    ),
    _(
        "pyb.Servo(id) \nCreate a servo object.\n\nParamaters:\nid (int) -- can be 1-4, and corresponds to pins X1 through X4.\n"
    ),
    _(
        "pyb.Servo.angle([angle, time=0]) \nIf arguments are given, sets the angle of the servo\nIf no arguments are given, gets the angle of the servo.\n\nParameters:\nangle (float or int) [optional] -- the angle to move to in degrees.\ntime (int) -- the number of milliseconds to take to get to the specified angle.\nIf omitted, then the servo moves as quickly as possible to its new position.\n\nReturns: the current angle of the servo (if no arguments are given)\n\nReturn type: float\n"
    ),
    _(
        "pyb.Servo.speed([speed, time=0]) \nIf no arguments are given, returns the current speed.\nIf arguments are given, sets the speed of the servo.\n\nParameters:\nspeed (int) -- the speed to change to, between -100 and 100.\ntime (int) -- the number of milliseconds to take to get to the specified speed.\nIf omitted, then the servo accelerates as quickly as possible.\n\nReturns: the current speed of the servo (if no arguments given)\n\nReturn type: int\n"
    ),
    _(
        "pyb.Servo.pulse_width([value]) \nIf no arguments are given, returns the current raw pulse-width value.\nIf an argument is given, sets the raw pulse-width value.\n\nParameters: value (int) [optional] -- the desired pulse width value.\n\nReturns: the current raw pulse-width value (if no arguments are given)\n\nReturn type: int\n"
    ),
    _(
        "pyb.SPI(bus, ...) \nConstruct an SPI object on the given bus. With no additional parameters, the SPI\nobject is created but not initialised. If extra arguments are given, the bus is\ninitialised. See init for parameters of initialisation.\n\nParameters:\nbus (int or string) -- can be 1 or 2, or ‘X’ or ‘Y’\nSPI(1) is on the X position: (NSS, SCK, MISO, MOSI) = (X5, X6, X7, X8) = (PA4, PA5, PA6, PA7)\nSPI(2) is on the Y position: (NSS, SCK, MISO, MOSI) = (Y5, Y6, Y7, Y8) = (PB12, PB13, PB14, PB15)\n"
    ),
    _("pyb.SPI.deinit() \nTurn off the SPI bus.\n"),
    _(
        "SPI.init(mode, baudrate=328125, *, prescaler, polarity=1, phase=0, bits=8, firstbit=SPI.MSB, ti=False, crc=None) \nInitialise the SPI bus with the given parameters\n\nParameters:\nmode (constant) -- must be either SPI.MASTER or SPI.SLAVE.\nbaudrate (int) -- the SCK clock rate (only sensible for a master).\nprescaler (int) -- the prescaler to use to derive SCK from the APB bus frequency, overrides baudrate.\npolarity (int) -- can be 0 or 1, the level the idle clock line sits at.\nphase (int) -- 0 or 1 to sample data on the first or second clock edge respectively.\nbits (int) -- 8 or 16, and is the number of bits in each transferred word.\nfirstbit (constant) -- can be SPI.MSB or SPI.LSB\ncrc (None) -- can be None for no CRC, or a polynomial specifier.\n"
    ),
    _(
        "pyb.SPI.recv(recv, *, timeout=5000) \nReceive data on the bus\n\nParameters:\nrecv (int or buffer) -- can be the number of bytes to receive, or a mutable\nbuffer, which will be filled with received bytes.\ntimeout (int) -- the timeout in milliseconds to wait for the receive.\n\nReturns: if recv is an integer then a new buffer of the bytes received, otherwise\nthe same buffer that was passed in to recv,\n\nReturn type: buffer\n"
    ),
    _(
        "pyb.SPI.send(send, *, timeout=5000) \nSend data on the bus\n\nParameters:\nsend (bytes) -- the data to send (an integer to send, or a buffer object)\ntimeout (int) -- the timeout in milliseconds to wait for the send\n"
    ),
    _(
        "pyb.SPI.send_recv(send, recv=None, *, timeout=5000) \nSend and receive data on the bus at the same time\n\nParameters:\nsend (bytes) -- the data to send (an integer to send, or a buffer object)\nrecv (buffer) -- a mutable buffer which will be filled with received bytes. It\ncan be the same as send, or omitted. If omitted, a new buffer will be created\ntimeout (int) -- the timeout in milliseconds to wait for the receive.\n\nReturns: buffer with the received bytes.\n\nReturn type: bytes\n"
    ),
    _(
        "pyb.Switch() \nCreate and return a switch object.\n\nReturns: Switch object\n\nReturn type: Switch class\n"
    ),
    _(
        "pyb.Switch.value() \nGet the current switch state.\n\nReturns: True if pressed down, False otherwise.\n\nReturn type: bool\n"
    ),
    _(
        "pyb.Switch.callback(fun) \nRegister the given function to be called when the switch is pressed down\n\nParameters:\nfun (function) -- the function to execute or None. If fun is None, then it disables the callback.\n"
    ),
    _(
        "pyb.Timer(id, ...) \nConstruct a new timer object of the given id. If additional arguments are given,\nthen the timer is initialised by init(...).\n\nParameters:\nid (int) -- can be 1 to 14.\n"
    ),
    _(
        "pyb.Timer.init(*, freq, prescaler, period) \nInitialise the timer. Initialisation must be either by frequency (in Hz) or by\nprescaler and period\n\nParameters:\nfreq (int) -- specifies the periodic frequency of the timer.\nprescaler (int) -- [0-0xffff] specifies the value to be loaded into the timer’s Prescaler\nRegister (PSC). The timer clock source is divided by prescaler + 1\nperiod (int) -- [0-0xffff] for timers 1, 3, 4, and 6-15. [0-0x3fffffff] for\ntimers 2 & 5. Specifies the value to be loaded into the timer’s AutoReload Register (ARR)\n"
    ),
    _(
        "pyb.Timer.deinit() \nDeinitialises the timer, callback, and channel callbacks associated with the Timer\n"
    ),
    _("pyb.Timer.counter([value]) \nGet or set the timer counter.\n"),
    _(
        "pyb.Timer.freq([value]) \nGet or set the frequency for the timer (changes prescaler and period if set).\n"
    ),
    _("pyb.Timer.period([value]) \nGet or set the period of the timer.\n"),
    _(
        "pyb.Timer.prescaler([value]) \nGet or set the prescaler for the timer.\n"
    ),
    _(
        "pyb,Timer.source_freq() \nGet the frequency of the source of the timer.\n"
    ),
    _(
        "pyb.UART(bus, ...) \nConstruct a UART object on the given bus.\nWith no additional parameters, the UART object is created but not initialised.\nIf extra arguments are given, the bus is initialised. See init for parameters\nof initialisation.\n\nParameters:\nbus (int or string) -- can be 1-6, or ‘XA’, ‘XB’, ‘YA’, or ‘YB’.\n"
    ),
    _(
        "pyb.UART.init(baudrate, bits=8, parity=None, stop=1, *, timeout=1000, flow=0, timeout_char=0, read_buf_len=64)\nInitialise the UART bus with the given parameters\n\nParameters:\nbaudrate (int) -- the clock rate\nbits (int) -- the number of bits per character, 7, 8 or 9\nparity (int) -- the parity, None, 0 (even) or 1 (odd)\nstop (int) -- the number of stop bits, 1 or 2\nflow (int) -- the flow control type. Can be 0, UART.RTS, UART.CTS or UART.RTS | UART.CTS\ntimeout (int) -- the timeout in milliseconds to wait for writing/reading the first character\ntimeout_char (int) -- the timeout in milliseconds to wait between characters\nwhile writing or reading.\nread_buf_len (int) -- the character length of the read buffer (0 to disable)\n"
    ),
    _("pyb.UART.deinit() \nTurn off the UART bus.\n"),
    _("pyb.UART.any() \nReturns the number of bytes waiting (may be 0).\n"),
    _(
        "pyb.UART.read([nbytes]) \nRead characters. If nbytes is specified then read at most that many bytes.\nIf nbytes is not given then the method reads as much data as possible. It returns\nafter the timeout has elapsed.\n\nParameters:\nnbytes (int) [optional] -- number of bytes to attempt to return\n\nReturns: a bytes object containing the bytes read in, None on timeout.\n\nReturn type: bytes or None"
    ),
    _(
        "pyb.UART.readchar() \nReceive a single character on the bus.\n\nReturns: the character read, as an integer, returns -1 on timeout.\n\nReturn type: int\n"
    ),
    _(
        "pyb.UART.readinto(buf[, nbytes]) \nRead bytes into the buf\n\nParameters:\nbuf (buffer) -- buffer to store bytes in\nnbytes (int) [optional] -- if specified then read at most that many bytes,\notherwise, read at most len(buf) bytes\n\nReturns: number of bytes read and stored into buf or None on timeout\n\nReturn type: int\n"
    ),
    _(
        "pyb.UART.readline() \nRead a line, ending in a newline character. If such a line exists, return is\nimmediate. If the timeout elapses, all available data is returned regardless of\nwhether a newline exists.\n\nReturns: the line read or None on timeout if no data is available\n\nReturn type: bytes\n"
    ),
    _(
        "pyb.UART.write(buf) \nWrite the buffer of bytes to the bus.\nParameters:\nbuf (buffer) -- if characters are 7 or 8 bits wide then each byte is one character. If characters\nare 9 bits wide then two bytes are used for each character (little endian), and\nbuf must contain an even number of bytes\n\nReturns: number of bytes written. If a timeout occurs and no bytes were written returns None\n\nReturn type: int\n"
    ),
    _(
        "pyb.UART.writechar(char) \nWrite a single character on the bus\n\nParameters:\nchar (int) -- the integer to write\n"
    ),
    _(
        "pyb.UART.sendbreak() \nSend a break condition on the bus.\nThis drives the bus low for a duration of 13 bits.\n"
    ),
    _("pyb.USB_HID() \nCreate a new USB_HID object.\n"),
    _(
        "pyb.USB_HID.recv(data, *, timeout=5000) \nReceive data on the bus\n\nParameters:\ndata can be an integer, which is the number of bytes to receive, or a mutable\nbuffer (int or buffer) -- which will be filled with received bytes.\ntimeout (int) -- the timeout in milliseconds to wait for the receive\n\nReturns: if data is an integer then a new buffer of the bytes received, otherwise\nthe number of bytes read into data is returned\n\nReturn type: int or buffer\n"
    ),
    _(
        "pyb.USB_HID.send(data) \nSend data over the USB HID interface\n\nParameters:\ndata (multiple) -- the data to send (a tuple/list of integers, or a bytearray)\n"
    ),
    _("pyb.USB_VCP() \nCreate a new USB_VCP object\n"),
    _(
        "pyb.USB_VCP.setinterrupt(chr) \nSet the character which interrupts running Python code, set to 3 (CTRL-C) by default.\n"
    ),
    _(
        "pyb.USB_VCP.isconnected() \nWhether USB is connected as a serial device\n\nReturns: True if USB is connected as a serial device, else False\n\nReturn type: bool\n"
    ),
    _(
        "pyb.USB_VCP.any() \nWhether any characters are waiting\n\nReturns: True if any characters waiting, else False\n\nReturn type: bool\n"
    ),
    _(
        "pyb.USB_VCP.close() \nThis method does nothing. It exists so the USB_VCP object can act as a file.\n"
    ),
    _(
        "pyb.USB_VCP.read([nbytes]) \nRead at most nbytes from the serial device and return them as a bytes object\n"
    ),
    _(
        "pyb.USB_VCP.readinto(buf[, maxlen]) \nRead bytes from the serial device and store them into buf\n\nParameters:\nbuf (buffer) -- a buffer-like object to read into\nmaxlen (int) -- if maxlen is given and then at most min(maxlen, len(buf)) bytes are read\n\nReturns: the number of bytes read and stored into buf or None if no pending data available\n\nReturn type: int\n"
    ),
    _(
        "pyb.USB_VCP.readline() \nRead a whole line from the serial device.\n\nReturns: a bytes object containing the data\n\nReturn type: bytes\n"
    ),
    _(
        "pyb.USB_VCP.readlines() \nRead as much data as possible from the serial device, breaking it into lines\n\nReturns: a list of bytes objects, each object being one of the lines\n\nReturn type: list\n"
    ),
    _(
        "pyb.USB_VCP.write(buf) \nWrite the bytes from buf to the serial device.\n"
    ),
    _(
        "pyb.USB_VCP.recv(data, *, timeout=5000) \nReceive data on the bus\n\nParameters:\ndata (int or buffer) -- the number of bytes to receive, or a mutable buffer,\nwhich will be filled with received bytes\ntimeout (int) -- the timeout in milliseconds to wait for the receive\n\nReturns: if data is an integer then a new buffer of the bytes received, otherwise\nthe number of bytes read into data is returned\nReturn type: int or buffer\n"
    ),
    _(
        "pyb.USB_VCP.send(data, *, timeout=5000) \nSend data over the USB VCP:\n\nParameters:\ndata (int of buffer) -- the data to send\ntimeout (int) -- the timeout in milliseconds to wait for the send.\n\nReturns: the number of bytes sent\n\nReturn type: int\n"
    ),
]



"""
UI related code for dialogs used by Mu.

Copyright (c) 2015-2017 Nicholas H.Tollervey and others (see the AUTHORS file).

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import logging
from PyQt5.QtCore import QSize, QProcess, QTimer, Qt

from PyQt5.QtWidgets import (
    QHBoxLayout,
    QVBoxLayout,
    QGridLayout,
    QListWidget,
    QLabel,
    QListWidgetItem,
    QDialog,
    QDialogButtonBox,
    QPlainTextEdit,
    QTabWidget,
    QWidget,
    QCheckBox,
    QLineEdit,
    QPushButton,
    QFileDialog,
    QGroupBox,
    QComboBox,
)
from PyQt5.QtGui import QTextCursor
from mu.resources import load_icon
from mu.interface.widgets import DeviceSelector
from ..virtual_environment import venv

logger = logging.getLogger(__name__)


class ModeItem(QListWidgetItem):
    """
    Represents an available mode listed for selection.
    """

    def __init__(self, name, description, icon, parent=None):
        super().__init__(parent)
        self.name = name
        self.description = description
        self.icon = icon
        text = "{}\n{}".format(name, description)
        self.setText(text)
        self.setIcon(load_icon(self.icon))


class ModeSelector(QDialog):
    """
    Defines a UI for selecting the mode for Mu.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

    def setup(self, modes, current_mode):
        self.setMinimumSize(600, 400)
        self.setWindowTitle(_("Select Mode"))
        widget_layout = QVBoxLayout()
        label = QLabel(
            _(
                'Please select the desired mode then click "OK". '
                'Otherwise, click "Cancel".'
            )
        )
        label.setWordWrap(True)
        widget_layout.addWidget(label)
        self.setLayout(widget_layout)
        self.mode_list = QListWidget()
        self.mode_list.itemDoubleClicked.connect(self.select_and_accept)
        widget_layout.addWidget(self.mode_list)
        self.mode_list.setIconSize(QSize(48, 48))
        for name, item in modes.items():
            if not item.is_debugger:
                litem = ModeItem(
                    item.name, item.description, item.icon, self.mode_list
                )
                if item.icon == current_mode:
                    self.mode_list.setCurrentItem(litem)
        self.mode_list.sortItems()
        instructions = QLabel(
            _(
                "Change mode at any time by clicking "
                'the "Mode" button containing Mu\'s logo.'
            )
        )
        instructions.setWordWrap(True)
        widget_layout.addWidget(instructions)
        button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        widget_layout.addWidget(button_box)

    def select_and_accept(self):
        """
        Handler for when an item is double-clicked.
        """
        self.accept()

    def get_mode(self):
        """
        Return details of the newly selected mode.
        """
        if self.result() == QDialog.Accepted:
            return self.mode_list.currentItem().icon
        else:
            raise RuntimeError("Mode change cancelled.")


class LogWidget(QWidget):
    """
    Used to display Mu's logs.
    """

    def setup(self, log):
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)
        label = QLabel(
            _(
                "When reporting a bug, copy and paste the content of "
                "the following log file."
            )
        )
        label.setWordWrap(True)
        widget_layout.addWidget(label)
        self.log_text_area = QPlainTextEdit()
        self.log_text_area.setReadOnly(True)
        self.log_text_area.setLineWrapMode(QPlainTextEdit.NoWrap)
        self.log_text_area.setPlainText(log)
        widget_layout.addWidget(self.log_text_area)


class EnvironmentVariablesWidget(QWidget):
    """
    Used for editing and displaying environment variables used with Python 3
    mode.
    """

    def setup(self, envars):
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)
        label = QLabel(
            _(
                "The environment variables shown below will be "
                "set each time you run a Python 3 script.\n\n"
                "Each separate enviroment variable should be on a "
                "new line and of the form:\nNAME=VALUE"
            )
        )
        label.setWordWrap(True)
        widget_layout.addWidget(label)
        self.text_area = QPlainTextEdit()
        self.text_area.setLineWrapMode(QPlainTextEdit.NoWrap)
        self.text_area.setPlainText(envars)
        widget_layout.addWidget(self.text_area)


class MicrobitSettingsWidget(QWidget):
    """
    Used for configuring how to interact with the micro:bit:

    * Minification flag.
    * Override runtime version to use.
    """

    def setup(self, minify, custom_runtime_path):
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)
        self.minify = QCheckBox(_("Minify Python code before flashing?"))
        self.minify.setChecked(minify)
        widget_layout.addWidget(self.minify)
        label = QLabel(
            _(
                "Override the built-in MicroPython runtime with "
                "the following hex file (empty means use the "
                "default):"
            )
        )
        label.setWordWrap(True)
        widget_layout.addWidget(label)
        self.runtime_path = QLineEdit()
        self.runtime_path.setText(custom_runtime_path)
        widget_layout.addWidget(self.runtime_path)
        widget_layout.addStretch()


class PackagesWidget(QWidget):
    """
    Used for editing and displaying 3rd party packages installed via pip to be
    used with Python 3 mode.
    """

    def setup(self, packages):
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)
        self.text_area = QPlainTextEdit()
        self.text_area.setLineWrapMode(QPlainTextEdit.NoWrap)
        label = QLabel(
            _(
                "The packages shown below will be available to "
                "import in Python 3 mode. Delete a package from "
                "the list to remove its availability.\n\n"
                "Each separate package name should be on a new "
                "line. Packages are installed from PyPI "
                "(see: https://pypi.org/)."
            )
        )
        label.setWordWrap(True)
        widget_layout.addWidget(label)
        self.text_area.setPlainText(packages)
        widget_layout.addWidget(self.text_area)


class LocaleWidget(QWidget):
    """
    Used for manually setting the locale (and thus the language) used by Mu.
    """

    LANGUAGES = {
        _("Automatically detect"): "",
        "English": "en",
        "Deutsch": "de_DE",
        "Español": "es",
        "Français": "fr",
        "日本語": "ja",
        "Polskie": "pl",
        "Português (Br)": "pt_BR",
        "Português (Pt)": "pt_PT",
        "русский язык": "ru_RU",
        "Slovenský": "sk_SK",
        "Svenska": "sv",
        "tiếng Việt": "vi",
        "中文": "zh_CN",
    }

    def setup(self, locale):
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)
        self.drop_down = QComboBox()
        for k, v in self.LANGUAGES.items():
            self.drop_down.addItem(k, v)
        index = self.drop_down.findData(locale)
        if index > -1:
            self.drop_down.setCurrentIndex(index)
        label = QLabel(
            _(
                "Please select the language for Mu's user interface from the "
                "choices listed below. <strong>Restart Mu for these changes "
                "to take effect.</strong>"
            )
        )
        label.setWordWrap(True)
        widget_layout.addWidget(label)
        widget_layout.addWidget(self.drop_down)
        widget_layout.addStretch()

    def get_locale(self):
        """
        Return the user-selected language code.
        """
        return self.LANGUAGES.get(self.drop_down.currentText(), "")


class ESPFirmwareFlasherWidget(QWidget):
    """
    Used for configuring how to interact with the ESP:

    * Override MicroPython.
    """

    def setup(self, mode, device_list):
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)

        # Instructions
        grp_instructions = QGroupBox(
            _("How to flash MicroPython to your device")
        )
        grp_instructions_vbox = QVBoxLayout()
        grp_instructions.setLayout(grp_instructions_vbox)
        # Note: we have to specify the link color here, to something
        # that's suitable for both day/night/contrast themes, as the
        # link color is not configurable in the Qt Stylesheets
        instructions = _(
            "&nbsp;1. Determine the type of device (ESP8266 or ESP32)<br />"
            "&nbsp;2. Download firmware from the "
            '<a href="https://micropython.org/download" '
            'style="color:#039be5;">'
            "https://micropython.org/download</a><br/>"
            "&nbsp;3. Connect your device<br/>"
            "&nbsp;4. Load the .bin file below using the 'Browse' button<br/>"
            "&nbsp;5. Press 'Erase & write firmware'"
            # "<br /><br />Check the current MicroPython version using the "
            # "following commands:<br />"
            # ">>> import sys<br />"
            # ">>> sys.implementation"
        )
        label = QLabel(instructions)
        label.setTextFormat(Qt.RichText)
        label.setTextInteractionFlags(Qt.TextBrowserInteraction)
        label.setOpenExternalLinks(True)
        label.setWordWrap(True)
        grp_instructions_vbox.addWidget(label)
        widget_layout.addWidget(grp_instructions)

        # Device type, firmware path, flash button
        device_selector_label = QLabel(_("Device:"))
        self.device_selector = DeviceSelector(show_label=True, icon_first=True)
        self.device_selector.set_device_list(device_list)
        device_type_label = QLabel(_("Choose device type:"))
        self.device_type = QComboBox(self)
        self.device_type.addItem("ESP8266")
        self.device_type.addItem("ESP32")
        firmware_label = QLabel(_("Firmware (.bin):"))
        self.txtFolder = QLineEdit()
        self.btnFolder = QPushButton(_("Browse"))
        self.btnExec = QPushButton(_("Erase && write firmware"))
        self.btnExec.setEnabled(False)
        form_set = QGridLayout()
        form_set.addWidget(device_selector_label, 0, 0)
        form_set.addWidget(self.device_selector, 0, 1, 1, 3)
        form_set.addWidget(device_type_label, 1, 0)
        form_set.addWidget(self.device_type, 1, 1)
        form_set.addWidget(firmware_label, 2, 0)
        form_set.addWidget(self.txtFolder, 2, 1)
        form_set.addWidget(self.btnFolder, 2, 2)
        form_set.addWidget(self.btnExec, 2, 3)
        widget_layout.addLayout(form_set)

        # Output area
        self.log_text_area = QPlainTextEdit()
        self.log_text_area.setReadOnly(True)
        form_set = QHBoxLayout()
        form_set.addWidget(self.log_text_area)
        widget_layout.addLayout(form_set)

        # Connect events
        self.txtFolder.textChanged.connect(self.firmware_path_changed)
        self.btnFolder.clicked.connect(self.show_folder_dialog)
        self.btnExec.clicked.connect(self.update_firmware)
        self.device_selector.device_changed.connect(self.toggle_exec_button)

        self.mode = mode

    def esptool_is_installed(self):
        """
        Is the 'esptool' package installed?
        """
        baseline, user_packages = venv.installed_packages()
        return "esptool" in user_packages

    def show_folder_dialog(self):
        # open dialog and set to foldername
        filename = QFileDialog.getOpenFileName(
            self,
            _("Select MicroPython firmware (.bin)"),
            os.path.expanduser("."),
            _("Firmware (*.bin)"),
        )
        if filename:
            filename = filename[0].replace("/", os.sep)
            self.txtFolder.setText(filename)

    def update_firmware(self):
        baudrate = 115200

        if self.mode.repl:
            self.mode.toggle_repl(None)
        if self.mode.plotter:
            self.mode.toggle_plotter(None)
        if self.mode.fs is not None:
            self.mode.toggle_files(None)

        device = self.device_selector.selected_device()
        if device is None:
            return

        esptool = "-mesptool"
        erase_command = '"{}" "{}" --port {} erase_flash'.format(
            venv.interpreter, esptool, device.port
        )

        if self.device_type.currentText() == "ESP32":
            write_command = (
                '"{}" "{}" --chip esp32 --port {} --baud {} '
                'write_flash -z 0x1000 "{}"'
            ).format(
                venv.interpreter,
                esptool,
                device.port,
                baudrate,
                self.txtFolder.text(),
            )
        else:
            write_command = (
                '"{}" "{}" --chip esp8266 --port {} --baud {} '
                'write_flash --flash_size=detect 0 "{}"'
            ).format(
                venv.interpreter,
                esptool,
                device.port,
                baudrate,
                self.txtFolder.text(),
            )

        self.commands = [erase_command, write_command]
        self.run_esptool()

    def run_esptool(self):
        self.process = QProcess(self)
        self.process.setProcessChannelMode(QProcess.MergedChannels)
        self.process.readyReadStandardError.connect(self.read_process)
        self.process.readyReadStandardOutput.connect(self.read_process)
        self.process.finished.connect(self.esptool_finished)
        self.process.error.connect(self.esptool_error)

        command = self.commands.pop(0)
        self.log_text_area.appendPlainText(command + "\n")
        self.process.start(command)

    def esptool_error(self, error_num):
        self.log_text_area.appendPlainText(
            "Error occurred: Error {}\n".format(error_num)
        )
        self.process = None

    def esptool_finished(self, exitCode, exitStatus):
        """
        Called when the subprocess that executes 'esptool.py is finished.
        """
        # Exit if a command fails
        if exitCode != 0 or exitStatus == QProcess.CrashExit:
            self.log_text_area.appendPlainText("Error on flashing. Aborting.")
            return
        if self.commands:
            self.process = None
            self.run_esptool()

    def read_process(self):
        """
        Read data from the child process and append it to the text area. Try
        to keep reading until there's no more data from the process.
        """
        msg = ""
        data = self.process.readAll()
        if data:
            try:
                msg = data.data().decode("utf-8")
                self.append_data(msg)
            except UnicodeDecodeError:
                pass
            QTimer.singleShot(2, self.read_process)

    def append_data(self, msg):
        """
        Add data to the end of the text area.
        """
        cursor = self.log_text_area.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(msg)
        cursor.movePosition(QTextCursor.End)
        self.log_text_area.setTextCursor(cursor)

    def firmware_path_changed(self):
        self.toggle_exec_button()

    def toggle_exec_button(self):
        if (
            len(self.txtFolder.text()) > 0
            and self.device_selector.selected_device() is not None
        ):
            self.btnExec.setEnabled(True)
        else:
            self.btnExec.setEnabled(False)


class AdminDialog(QDialog):
    """
    Displays administrative related information and settings (logs, environment
    variables, third party packages etc...).
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.microbit_widget = None
        self.package_widget = None
        self.envar_widget = None

    def setup(self, log, settings, packages, mode, device_list):
        self.setMinimumSize(600, 400)
        self.setWindowTitle(_("Mu Administration"))
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)
        self.tabs = QTabWidget()
        widget_layout.addWidget(self.tabs)
        button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        widget_layout.addWidget(button_box)
        # Tabs
        self.log_widget = LogWidget(self)
        self.log_widget.setup(log)
        self.tabs.addTab(self.log_widget, _("Current Log"))
        if mode.short_name in ["python", "web", "pygamezero"]:
            self.envar_widget = EnvironmentVariablesWidget(self)
            self.envar_widget.setup(settings.get("envars", ""))
            self.tabs.addTab(self.envar_widget, _("Python3 Environment"))
        if mode.short_name == "microbit":
            self.microbit_widget = MicrobitSettingsWidget(self)
            self.microbit_widget.setup(
                settings.get("minify", False),
                settings.get("microbit_runtime", ""),
            )
            self.tabs.addTab(self.microbit_widget, _("BBC micro:bit Settings"))
        if mode.short_name in ["python", "web", "pygamezero"]:
            self.package_widget = PackagesWidget(self)
            self.package_widget.setup(packages)
            self.tabs.addTab(self.package_widget, _("Third Party Packages"))
        if mode.short_name == "esp":
            self.esp_widget = ESPFirmwareFlasherWidget(self)
            self.esp_widget.setup(mode, device_list)
            self.tabs.addTab(self.esp_widget, _("ESP Firmware flasher"))
        # Configure local.
        self.locale_widget = LocaleWidget(self)
        self.locale_widget.setup(settings.get("locale"))
        self.tabs.addTab(
            self.locale_widget, load_icon("language.svg"), _("Select Language")
        )
        self.log_widget.log_text_area.setFocus()

    def settings(self):
        """
        Return a dictionary representation of the raw settings information
        generated by this dialog. Such settings will need to be processed /
        checked in the "logic" layer of Mu.
        """
        settings = {}
        if self.envar_widget:
            settings["envars"] = self.envar_widget.text_area.toPlainText()
        if self.microbit_widget:
            settings["minify"] = self.microbit_widget.minify.isChecked()
            settings[
                "microbit_runtime"
            ] = self.microbit_widget.runtime_path.text()
        if self.package_widget:
            settings["packages"] = self.package_widget.text_area.toPlainText()
        settings["locale"] = self.locale_widget.get_locale()
        return settings


class FindReplaceDialog(QDialog):
    """
    Display a dialog for getting:

    * A term to find,
    * An optional value to replace the search term,
    * A flag to indicate if the user wishes to replace all.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

    def setup(self, find=None, replace=None, replace_flag=False):
        self.setMinimumSize(600, 200)
        self.setWindowTitle(_("Find / Replace"))
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)
        # Find.
        find_label = QLabel(_("Find:"))
        self.find_term = QLineEdit()
        self.find_term.setText(find)
        self.find_term.selectAll()
        widget_layout.addWidget(find_label)
        widget_layout.addWidget(self.find_term)
        # Replace
        replace_label = QLabel(_("Replace (optional):"))
        self.replace_term = QLineEdit()
        self.replace_term.setText(replace)
        widget_layout.addWidget(replace_label)
        widget_layout.addWidget(self.replace_term)
        # Global replace.
        self.replace_all_flag = QCheckBox(_("Replace all?"))
        self.replace_all_flag.setChecked(replace_flag)
        widget_layout.addWidget(self.replace_all_flag)
        button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        widget_layout.addWidget(button_box)

    def find(self):
        """
        Return the value the user entered to find.
        """
        return self.find_term.text()

    def replace(self):
        """
        Return the value the user entered for replace.
        """
        return self.replace_term.text()

    def replace_flag(self):
        """
        Return the value of the global replace flag.
        """
        return self.replace_all_flag.isChecked()


class PackageDialog(QDialog):
    """
    Display the output of the pip commands needed to remove or install
    packages.

    Because the QProcess mechanism we're using is asynchronous, we have to
    manage the pip requests via `pip_queue`. When one request is signalled
    as finished we start the next.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.pip_queue = []

    def setup(self, to_remove, to_add):
        """
        Create the UI for the dialog.
        """
        # Basic layout.
        self.setMinimumSize(600, 400)
        self.setWindowTitle(_("Third Party Package Status"))
        widget_layout = QVBoxLayout()
        self.setLayout(widget_layout)
        # Text area for pip output.
        self.text_area = QPlainTextEdit()
        self.text_area.setReadOnly(True)
        self.text_area.setLineWrapMode(QPlainTextEdit.NoWrap)
        widget_layout.addWidget(self.text_area)
        # Buttons.
        self.button_box = QDialogButtonBox(QDialogButtonBox.Ok)
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(False)
        self.button_box.accepted.connect(self.accept)
        widget_layout.addWidget(self.button_box)

        #
        # Set up the commands to be issues to pip. Since we'll be popping
        # from the list (as LIFO) we'll add the installs first so the
        # removes are the first to happen
        #
        if to_add:
            self.pip_queue.append(("install", to_add))
        if to_remove:
            self.pip_queue.append(("remove", to_remove))
        QTimer.singleShot(2, self.next_pip_command)

    def next_pip_command(self):
        """
        Run the next pip command, finishing if there is none.
        """
        if self.pip_queue:
            command, packages = self.pip_queue.pop()
            self.run_pip(command, packages)
        else:
            self.finish()

    def finish(self):
        """
        Set the UI to a valid end state.
        """
        self.text_area.appendPlainText("\nFINISHED")
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(True)

    def run_pip(self, command, packages):
        """
        Run a pip command in a subprocess and pipe the output to the dialog's
        text area.
        """
        if command == "remove":
            pip_fn = venv.remove_user_packages
        elif command == "install":
            pip_fn = venv.install_user_packages
        else:
            raise RuntimeError(
                "Invalid pip command: %s %s" % (command, packages)
            )
        pip_fn(
            packages,
            slots=venv.Slots(
                output=self.text_area.appendPlainText,
                finished=self.next_pip_command,
            ),
        )



"""
A mode for working with ESP8266 and ESP32 boards running MicroPython.

Copyright (c) 2015-2019 Nicholas H.Tollervey and others (see the AUTHORS file).

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import logging
from mu.modes.base import MicroPythonMode, FileManager
from mu.modes.api import ESP_APIS, SHARED_APIS
from mu.interface.panes import CHARTS
from PyQt5.QtCore import QThread
import os


logger = logging.getLogger(__name__)


class ESPMode(MicroPythonMode):
    """
    Represents the functionality required for running MicroPython on ESP8266
    """

    name = _("ESP MicroPython")
    short_name = "esp"
    board_name = "ESP8266/ESP32"
    description = _("Write MicroPython on ESP8266/ESP32 boards.")
    icon = "esp"
    fs = None

    # The below list defines the supported devices, however, many
    # devices are using the exact same FTDI USB-interface, with vendor
    # ID 0x403 without reporting their own VID/PID

    # In some instances we can recognize the device not on VID/PID,
    # but on manufacturer ID, that's what the third column is for.
    # These more specific device specifications, should be listed
    # before the generic FTDI VID/PID's
    valid_boards = [
        # VID  , PID,    Manufacturer string, Device name
        (0x1A86, 0x7523, None, "HL-340"),
        (0x10C4, 0xEA60, None, "CP210x"),
        (0x0403, 0x6001, "M5STACK Inc.", "M5Stack ESP32 device"),
        (0x0403, 0x6001, None, None),  # FT232/FT245 (XinaBox CW01, CW02)
        (0x0403, 0x6010, None, None),  # FT2232C/D/L/HL/Q (ESP-WROVER-KIT)
        (0x0403, 0x6011, None, None),  # FT4232
        (0x0403, 0x6014, None, None),  # FT232H
        (0x0403, 0x6015, None, None),  # FT X-Series (Sparkfun ESP32)
        (0x0403, 0x601C, None, None),  # FT4222H
    ]

    def actions(self):
        """
        Return an ordered list of actions provided by this module. An action
        is a name (also used to identify the icon) , description, and handler.
        """
        buttons = [
            {
                "name": "run",
                "display_name": _("Run"),
                "description": _(
                    "Run your code directly on the {board_name}"
                    " via the REPL."
                ).format(board_name=self.board_name),
                "handler": self.run,
                "shortcut": "F5",
            },
            {
                "name": "files",
                "display_name": _("Files"),
                "description": _(
                    "Access the file system on {board_name}."
                ).format(board_name=self.board_name),
                "handler": self.toggle_files,
                "shortcut": "F4",
            },
            {
                "name": "repl",
                "display_name": _("REPL"),
                "description": _(
                    "Use the REPL to live-code on the {board_name}."
                ).format(board_name=self.board_name),
                "handler": self.toggle_repl,
                "shortcut": "Ctrl+Shift+I",
            },
        ]
        if CHARTS:
            buttons.append(
                {
                    "name": "plotter",
                    "display_name": _("Plotter"),
                    "description": _("Plot incoming REPL data."),
                    "handler": self.toggle_plotter,
                    "shortcut": "CTRL+Shift+P",
                }
            )
        return buttons

    def api(self):
        """
        Return a list of API specifications to be used by auto-suggest and call
        tips.
        """
        return SHARED_APIS + ESP_APIS

    def toggle_repl(self, event):
        if self.fs is None:
            if self.repl:
                # Remove REPL
                super().toggle_repl(event)
                self.set_buttons(files=True)
            elif not (self.repl):
                # Add REPL
                super().toggle_repl(event)
                if self.repl:
                    self.set_buttons(files=False)
        else:
            message = _("REPL and file system cannot work at the same time.")
            information = _(
                "The REPL and file system both use the same USB "
                "serial connection. Only one can be active "
                "at any time. Toggle the file system off and "
                "try again."
            )
            self.view.show_message(message, information)

    def toggle_plotter(self, event):
        """
        Check for the existence of the file pane before toggling plotter.
        """
        if self.fs is None:
            super().toggle_plotter(event)
            if self.plotter:
                self.set_buttons(files=False)
            elif not (self.repl or self.plotter):
                self.set_buttons(files=True)
        else:
            message = _(
                "The plotter and file system cannot work at the same " "time."
            )
            information = _(
                "The plotter and file system both use the same "
                "USB serial connection. Only one can be active "
                "at any time. Toggle the file system off and "
                "try again."
            )
            self.view.show_message(message, information)

    def run(self):
        """
        Takes the currently active tab, compiles the Python script therein into
        a hex file and flashes it all onto the connected device.
        """
        """
        if self.repl:
            message = _("Flashing cannot be performed at the same time as the "
                        "REPL is active.")
            information = _("File transfers use the same "
                            "USB serial connection as the REPL. Toggle the "
                            "REPL off and try again.")
            self.view.show_message(message, information)
            return
        """
        logger.info("Running script.")
        # Grab the Python script.
        tab = self.view.current_tab
        if tab is None:
            # There is no active text editor.
            message = _("Cannot run anything without any active editor tabs.")
            information = _(
                "Running transfers the content of the current tab"
                " onto the device. It seems like you don't have "
                " any tabs open."
            )
            self.view.show_message(message, information)
            return
        python_script = tab.text().split("\n")
        if not self.repl:
            self.toggle_repl(None)
        if self.repl and self.connection:
            self.connection.send_commands(python_script)

    def toggle_files(self, event):
        """
        Check for the existence of the REPL or plotter before toggling the file
        system navigator for the MicroPython device on or off.
        """
        if self.repl:
            message = _(
                "File system cannot work at the same time as the "
                "REPL or plotter."
            )
            information = _(
                "The file system and the REPL and plotter "
                "use the same USB serial connection. Toggle the "
                "REPL and plotter off and try again."
            )
            self.view.show_message(message, information)
        else:
            if self.fs is None:
                self.add_fs()
                if self.fs:
                    logger.info("Toggle filesystem on.")
                    self.set_buttons(run=False, repl=False, plotter=False)
            else:
                self.remove_fs()
                logger.info("Toggle filesystem off.")
                self.set_buttons(run=True, repl=True, plotter=True)

    def add_fs(self):
        """
        Add the file system navigator to the UI.
        """

        # Find serial port the ESP8266/ESP32 is connected to
        device = self.editor.current_device

        # Check for MicroPython device
        if not device:
            message = _("Could not find an attached {board_name}").format(
                board_name=self.board_name
            )
            information = _(
                "Please make sure the device is plugged "
                "into this computer.\n\nThe device must "
                "have MicroPython flashed onto it before "
                "the file system will work.\n\n"
                "Finally, press the device's reset button "
                "and wait a few seconds before trying "
                "again."
            )
            self.view.show_message(message, information)
            return
        self.file_manager_thread = QThread(self)
        self.file_manager = FileManager(device.port)
        self.file_manager.moveToThread(self.file_manager_thread)
        self.file_manager_thread.started.connect(self.file_manager.on_start)

        # Show directory of the current file in the left pane, if any,
        # otherwise show the default workspace_dir
        if self.view.current_tab and self.view.current_tab.path:
            path = os.path.dirname(os.path.abspath(self.view.current_tab.path))
        else:
            path = self.workspace_dir()
        self.fs = self.view.add_filesystem(
            path,
            self.file_manager,
            _("{board_name} board").format(board_name=self.board_name),
        )
        self.fs.set_message.connect(self.editor.show_status_message)
        self.fs.set_warning.connect(self.view.show_message)
        self.file_manager_thread.start()

    def remove_fs(self):
        """
        Remove the file system navigator from the UI.
        """
        self.view.remove_filesystem()
        self.file_manager = None
        self.file_manager_thread = None
        self.fs = None

    def on_data_flood(self):
        """
        Ensure the Files button is active before the REPL is killed off when
        a data flood of the plotter is detected.
        """
        self.set_buttons(files=True)
        super().on_data_flood()

    def deactivate(self):
        """
        Invoked whenever the mode is deactivated.
        """
        super().deactivate()
        if self.fs:
            self.remove_fs()

    def device_changed(self, new_device):
        """
        Invoked when the user changes device.
        """
        super().device_changed(new_device)
        if self.fs:
            self.remove_fs()
            self.add_fs()



"""
Contains definitions for Lego Spike APIs so they can be used in the editor for
autocomplete and call tips.

Copyright (c) 2015-2021 Nicholas H.Tollervey and others (see the AUTHORS file).

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

LEGO_APIS = [
    _(
        "battery.capacity_left() \nReturns the capacity left on the battery as a percentage of fully charged.\n"
    ),
    _(
        "battery.charger_detect() \nReturns the charger detected.\nReturns False if  no charger is attached.\nIf a charger has been attached the following is returned:\n\n\nflipper.USB_CH_PORT_NONE = 0\nflipper.USB_CH_PORT_SDP  = 1\nflipper.USB_CH_PORT_CDP  = 2\nflipper.USB_CH_PORT_DCP  = 3\n\n\nSDP is a Standard Downstream Port (typical USB port)\nDCP is a Dedicated Charging Port (high current USB port)\nCDP is a Charging Downstream Port (wall charger)\n"
    ),
    _(
        "battery.current() \nReturn the battery current consumption in mA. This is the current flowing from the battery.\n"
    ),
    _(
        "battery.info() \nReturns a dict with the Battery subsystem information. Avoid calling\nthis function every time you need a specifc item. It’s better to save\nthe result in a local variable and then reference the dict elements as\nneeded:\n\nerror_state - A list of currently active errors incl. 0 (flipper.BATTERY_NO_ERROR) if no errors are active. Possible error are:\n\n\nbattery.BATTERY_NO_ERROR = 0\nbattery.BATTERY_HUB_TEMPERATURE_CRITICAL_OUT_OF_RANGE = -1\nbattery.BATTERY_TEMPERATURE_OUT_OF_RANGE = -2\nbattery.BATTERY_TEMPERATURE_SENSOR_FAIL = -3\nbattery.BATTERY_BAD_BATTERY = -4\nbattery.BATTERY_VOLTAGE_TOO_LOW = -5\n\n\n\ncharger_state - The state of the charter circuit:\n\n\nbattery.DISCHARGING = 0\nbattery.CHARGING_ONGOING = 1\nbattery.CHARGING_COMPLETED = 2\nbattery.FAIL = -1\n\n\n\ncharge_voltage - The battery volgate in mV\n\ncharge_voltage_filtered - The battery volgate (filtered) in mV\n\ntemperature - The battery temperature in degree Celcius.\n\ncharge_current - The battery charging current in mA (the current flowing into the battery)\n\n\nNote\nThe returned value is not calibrated.\n\n\n\nbattery_capacity_left - The battery capacity left in percent of full capacity\n\n\n"
    ),
    _(
        "battery.temperature() \nReturn the battery temperature in degree Celcius.\n"
    ),
    _("battery.voltage() \nReturn the battery voltage in millivolts.\n"),
    _("hub.BT_VCP() \nCreate a new BT_VCP object.\n\n"),
    _(
        "hub.Button() \nRepresents the set of methods that can be called on any of the 4 buttons\non the Hub.\n\n"
    ),
    _(
        "hub.Image(string) \nIf string is used, it has to consist of digits 0-9 arranged into\nlines, describing the image, for example:\n\nor:\n\nwill create a 5×5 image of an X. The end of a line is indicated by a colon.\nIt’s also possible to use a newline (n) to indicate the end of a line\nlike this:\n\nor:\n\nThe other form creates an empty image with width columns and\nheight rows. Optionally buffer can be an array of\nwidth``×``height integers in range 0-9 to initialize the image, for example:\n\nor:\n\n"
    ),
    _(
        "hub.Image.get_pixel(x, y) \nReturn the brightness of pixel at column x and row y as an\ninteger between 0 and 9.\n"
    ),
    _("hub.Image.height() \nReturn the numbers of rows in the image.\n"),
    _(
        "hub.Image.set_pixel(x, y, value) \nSet the brightness of the pixel at column x and row y to the\nvalue, which has to be between 0 (dark) and 9 (bright).\nThis method will raise an exception when called on any of the built-in\nread-only images, like Image.HEART.\n"
    ),
    _("hub.Image.width() \nReturn the number of columns in the image.\n"),
    _("hub.Led() \nRepresents the center RGB LED.\n\n"),
    _(
        "hub.MotorPair() \nRepresents the set of methods that can be called on a motor pair\nobject after it is created using the hub.port.X.motor.pair()\n\n"
    ),
    _(
        "hub.MotorPair.brake() \nForce the motor driver to the Brake state. The motor driver shorts its output\nterminals and the motor will stop more quickly than if it Floats. Note\nthat this is NOT the same as actively holding motor position.\n"
    ),
    _(
        "hub.MotorPair.callback(fun(v)) \nA callback function that is invoked when the command is interrupted or completes\nThe callback function accepts one and only one parameter that indicates why the callback was\ninitiated. The following reasons can apply:\n\n0 = MOTOR_EVENT_COMPLETED\n1 = MOTOR_EVENT_INTERRUPTED\n2 = MOTOR_EVENT_STALL\n\n"
    ),
    _(
        "hub.MotorPair.float() \nForce the motor driver to the Float state. The motor driver no longer delivers\na PWM signal and the motor will coast to a stop if it is running.\n"
    ),
    _(
        "hub.MotorPair.hold() \nForce the motor driver to the Brake state. The motor will actively hold its target position.\n"
    ),
    _("hub.MotorPair.id() \nReturns the paired motor ID.\n"),
    _(
        "hub.MotorPair.pid() \nWith no parameters returns a tuple with the current used P, I and D values\nif the values have been set using this function or motor.default(). If not\nset the values returned is 0 (zero) and is invalid.\n\nNote\nAt the moment it is not possible to readout the default PID values used\nin the low-level drivers. To do this it is required to implement additional\nsub-commands in the LPF2 protocol.\n\n"
    ),
    _(
        "hub.MotorPair.preset(position0, position1) \nPresets the motor encoder with a new position value.\n\n‘position0’ - [0 - 4294967296] - Encoder value.\n‘position1’ - [0 - 4294967296] - Encoder value.\n\n"
    ),
    _(
        "hub.MotorPair.primary() \nReturns the motor object that initiated the pair() operation.\n"
    ),
    _(
        "hub.MotorPair.pwm(value0, value1) \n\nvalue [-100 - 100] Applies a PWM signal to the power pins of the port.\n\nUseful for controlling directly. The polarity of the PWM signal matches the\nsign of the value.\nA value of zero stops the PWM signal and leaves the motor driver in\nthe floating state.\n"
    ),
    _(
        "hub.MotorPair.run_at_speed(speed0, speed1, max_power, acceleration, deceleration) \nKeyword arguments:\n\nspeed [-100 - 100] - specifies the speed of the motor.\nmax_power [0 - 100] - specifies the maximum power of the motor when regulating the speed\nacceleration [0 - 10000] - specifies the time in msec for the motor to achieve specified speed\ndeceleration [0 - 10000] - specifies the time in msec for the motor to come to stop from full speed\n\n"
    ),
    _(
        "hub.MotorPair.run_for_degrees(degrees, speed0, speed1, max_power, acceleration, deceleration, stop) \nKeyword arguments:\n\ndegrees [-MAXINT - +MAXINT] - specifies the degrees to turn from current target position\nspeed [-100 - 100] - specifies the speed of the motor.\nmax_power [0 - 100] - specifies the maximum power of the motor when regulating the speed\nacceleration [0 - 10000] - specifies the time in msec for the motor to achieve specified speed\ndeceleration [0 - 10000] - specifies the time in msec for the motor to come to stop from full speed\nstop - specifies the stop state of the motor. [0 = MOTOR_STOP_FLOAT, 1 = MOTOR_STOP_BRAKE, 2 = MOTOR_STOP_HOLD]\n\n"
    ),
    _(
        "hub.MotorPair.run_for_time(msec, speed0, speed1, max_power, acceleration, deceleration, stop) \n\nmsec [0 - 65535] - Force the motor driver to run for msec milliseconds at speed. When the msec time expires the motor driver is forced to the stop state. If msec is negative a time of 0 (zero) milliseconds is assumed.\nmsec [-100 - 10000] - specifies the amount of milliseconds.\nspeed [-100 - 100] - specifies the speed of the motor.\nmax_power [0 - 100] - specifies the maximum power of the motor when regulating the speed\nacceleration [0 - 10000] - specifies the time in msec for the motor to achieve specified speed\ndeceleration [0 - 10000] - specifies the time in msec for the motor to come to stop from full speed\nstop - specifies the stop state of the motor. See ‘run_for_degrees’ method for stop types.\n\n"
    ),
    _(
        "hub.MotorPair.run_to_position(position0, position1, speed, max_power, acceleration, deceleration, stop) \nKeyword arguments:\n\nposition [-MAXINT - +MAXINT] - specifies the position of the motor.\nspeed [-100 - 100] - specifies the speed of the motor.\nmax_power [0 - 100] - specifies the maximum power of the motor when regulating the speed\nacceleration [0 - 10000] - specifies the time in msec for the motor to achieve specified speed\ndeceleration [0 - 10000] - specifies the time in msec for the motor to come to stop from full speed\nstop - specifies the stop state of the motor. See ‘run_for_degrees’ method for stop types.\n\n"
    ),
    _(
        "hub.MotorPair.secondary() \nReturns the motor object that was the parameter in the pair() operation.\n"
    ),
    _(
        "hub.MotorPair.unpair() \nUndo a pair of two motors. After call of this method the <MotorPair> object\nis invalid.\nThis method will return True if the uppair was successful. Otherwize False\nis returned (if a timeout has happend).\n"
    ),
    _("hub.Sound() \nRepresents a sound generator.\n\n"),
    _(
        "hub.Sound.beep(100-10000, 0-32767, 0-3) \nPlay a beep tone on the sound system. The parameters are all optional, and\nin order are:\n\n\nfrequency: defaults to 1000 Hz\ntime: defaults to 1000 msec\nwaveform: defaults to 0 (see below)\n\n\nThere are 4 different waveforms available:\n\n\n0 = sin\n1 = square\n2 = triangle\n3 = sawtooth\n\n\n"
    ),
    _(
        "hub.Sound.callback(fun(v)) \nRegister the given function to be called when the sound completes\nor is interrupted. If fun is None, then the callback is disabled.\nThe function accepts one of the following values:\n0 if the sound completed playing\n1 if the sound was interrupted\n"
    ),
    _("hub.Sound.off() \nTurns off the internal amplifier chip.\n"),
    _(
        "hub.Sound.on() \nTurns on the internal amplifier chip. Default at power-up.\n"
    ),
    _(
        "hub.Sound.play(filename, 12000-20000) \nPlay a sound file at the current volume. The filename must be specified, if\nthe file does not exist no sound is played.\nThe sound file must be raw 8 bit data at 16 kHz, and the second optional\nparameter specifies the playback speed. Currently the range is 12-20 kHz\nbut we may chnage that to +/- 10 or some other reasonable value.\n"
    ),
    _(
        "hub.Sound.volume(0-10) \nGet or set the volume of the sound system\n\nWith no argument, return 0-10 depending on the current volume\nWith volume given, set the volume of the sound system\n\n"
    ),
    _("hub.USB_VCP() \nCreate a new USB_VCP object.\n\n"),
    _("hub.battery() \n"),
    _("hub.display() \nCreate and return a display object.\n\n"),
    _(
        "port.X() \nRepresents the set of methods that can be called on any of the 6 ports\non the Hub.\n\n"
    ),
    _(
        "port.X.callback(fun) \nSet the function to be called when t hotplug event occurs on the port.\nIf fun is None then the callback will be disabled.\nfun() should take one parameter that indicates why the callback was\ninitiated. The following reasons can apply:\n\n0 = PORT_EVENT_DETACHED\n1 = PORT_EVENT_ATTACHED\n\n"
    ),
    _(
        "port.X.device() \nRepresents the set of methods that can be called on a device plugged\ninto any of the 6 ports on the Hub.\n\n"
    ),
    _(
        "port.X.device.get() \nReturns a list of value(s) that the currently active device mode makes\navailable. A device can be in single or combi-mode. In either case\nthe return value is a list of one or more values representing the\ndata described in the corresponing mode command.\nThere are 3 different formats available:\n\n\n0 = Raw\n1 = Pct\n2 = SI\n\n\n"
    ),
    _(
        "port.X.device.mode() \nPuts the device in the specified mode(s) depending on the specified \nmode value(s).\nIf the mode specifier is an integer, then the device is put into that mode.\nThe data returned by get() is one or more values corresponding to the\nnumber of datasets available in that mode. For example, a color sensor in\nindex color mode returns exactly one value, but in RGB mode a list of three\nvalues is returned.\nIf the mode specifier is a list, then we are asking for the device to be put\ninto combi-mode. The get() function now returns exactly the values corresponding\nto the requested modes and the datasets within that mode.\n\nNote\nThe mode specifier must be a list of 2-element tuples. If one of the list\nelements is not a 2-element tuple then it is ignored without error or warning.\n\nIf the mode specifier is an integer and it is followed by an argument that contains\na byte array - this array will be sent to the device as an output mode write\n"
    ),
    _(
        "port.X.device.pwm(value) \nKeyword arguments:\n\nvalue [-100 - 100] Applies a PWM signal to the power pins of the port.\n\nUseful for controlling directly. The polarity of the PWM signal matches the\nsign of the value.\nA value of zero stops the PWM signal and leaves the port driver in\nthe floating state.\n"
    ),
    _(
        "port.X.info() \nReturns a dictionary describing the capabilities of the device connected\nto the port. If a port has nothing plugged in, then the result is a\ndictionary with only a type key with a value of None.\nA port with a PoweredUp compatible device plugged in returns a dictionary\nlike this:\n\nfw_version : Firmware version as a 32 bit unsigned integer\nhw_version : Hardware version as a 32 bit unsigned integer\ntype : PoweredUp device type as an integer\ncombi_modes : A list of legal combi-modes as 16 bit unsigend integers\nspeed : The maximum baud rate of the device (0 for simple IDs)\nmodes : A list of dictionaries representing available modes\n\nEach modes list item dictionary has the  following keys:\n\nname : The name of the mode as a string\ncapability : The 48 capability bits as a binary string of 6 characters\nsymbol : The SI ymbol name of the data returned by the device in SI format\nraw : The min and max range of raw data expressed as a 2 element tuple\npct : The min and max range of % range data expressed as a 2 element tuple\nsi : The min and max range of si data expressed as a 2 element tuple\nmap_out : The output mapping bits as an 8 bit value\nmap_in : The infou mapping bits as an 8 bit value\nformat : A dictionary representing the format data for this mode\n\nEach format dictionary has the  following keys:\n\ndatasets : The number of data values that this mode returns\nfigures : The number of digits in the data value\ndecimals : The number of digits after the implied decimal point\ntype : The type of return data (signed 8, 16, 32 or float)\n\n"
    ),
    _("port.X.mode(mode) \nSet the mode of the port.\n"),
    _(
        "port.X.motor() \nRepresents the set of methods that can be called on a motor plugged\ninto any of the 6 ports on the Hub.\n\n"
    ),
    _(
        "port.X.motor.brake() \nForce the motor driver to the Brake state. The motor driver shorts its output\nterminals and the motor will stop more quickly than if it Floats. Note\nthat this is NOT the same as actively holding motor position.\n"
    ),
    _(
        "port.X.motor.busy(type) \nCheck to see if the motor is busy.\nKeyword argument(s):\n\n‘type’ - specifies which operation type to check for: mode/sensor (BUSY_MODE) or motor (BUSY_MOTOR).\n\n"
    ),
    _(
        "port.X.motor.callback(fun(v)) \nA callback function that is invoked when the command is interrupted or completes\nThe callback function accepts one and only one parameter that indicates why the callback was\ninitiated. The following reasons can apply:\n\n0 = MOTOR_EVENT_COMPLETED\n1 = MOTOR_EVENT_INTERRUPTED\n2 = MOTOR_EVENT_STALL\n\n"
    ),
    _(
        "port.X.motor.default() \nWith no parameters, returns a dict with the current default values that are\nused for motor operations if the corresponding keyword arguments are not specified.\nThe current defaults may be:\n\nspeed - Not implemented\nmax_power - Max % pwm to be used in regulated speed commands\nacceleration - Time in msec to reach 100% of motor design speed from 0%\ndeceleration - Time in msec to reach 0% of motor design speed from 100%\nstop - Default stop action at end of comment\npid - Default PID value used if set here or by motor.pid(). If not set the values returned is 0 (zero)\nstall - Stall detection True/False.\ncallback - A callback function that is invoked when the command is interrupted or completes\n\nThe callback function accepts one and only one parameter that indicates why the callback was\ninitiated. The following reasons can apply:\n\n0 = MOTOR_EVENT_COMPLETED\n1 = MOTOR_EVENT_INTERRUPTED\n2 = MOTOR_EVENT_STALL\n\n\nNote\nAt the moment it is not possible to readout the default PID values used\nin the low-level drivers. To do this it is required to implement additional\nsub-commands in the LPF2 protocol.\n\n"
    ),
    _(
        "port.X.motor.float() \nForce the motor driver to the Float state. The motor driver no longer delivers\na PWM signal and the motor will coast to a stop if it is running.\n"
    ),
    _(
        "port.X.motor.get() \nReturns the value(s) that the currently active device mode makes\navailable.\n\nNote\nSee the detailed explanation in the port.X.device section\n\n"
    ),
    _(
        "port.X.motor.hold() \nForce the motor driver to the Brake state. The motor will actively hold its target position.\n"
    ),
    _(
        "port.X.motor.mode(value) \nPuts the device in the specified mode.\n\nNote\nSee the detailed explanation in the port.X.device section\n\n"
    ),
    _(
        "port.X.motor.pair(motor) \nCreate a <MotorPair> object. The motor parameter must be motor object\nand cannot be the motor object that the pair() method is called on.\nIf the <MotorPair> object cannot be created due to a timeout False is returned.\nIt the object could not be created due to another fault None is retuned.\nSee the MotorPair API for methods that can be applied to a motor pair.\n"
    ),
    _(
        "port.X.motor.pid() \nWith no parameters returns a tuple with the current used P, I and D values\nif the values have been set using this function or motor.default(). If not\nset the values returned is 0 (zero) and is invalid.\n\nNote\nAt the moment it is not possible to readout the default PID values used\nin the low-level drivers. To do this it is required to implement additional\nsub-commands in the LPF2 protocol.\n\n"
    ),
    _(
        "port.X.motor.preset(position) \nPresets the motor encoder with a new position value.\n\n‘position’ - [0 - 4294967295] - Encoder value.\n\n"
    ),
    _(
        "port.X.motor.pwm(value) \nKeyword arguments:\n\nvalue [-100 - 100] Applies a PWM signal to the power pins of the port.\n\nUseful for controlling directly. The polarity of the PWM signal matches the\nsign of the value.\nA value of zero stops the PWM signal and leaves the motor driver in\nthe floating state.\n"
    ),
    _(
        "port.X.motor.run_at_speed(speed, max_power, acceleration, deceleration, stall) \nKeyword arguments:\n\nspeed [-100 - 100] - specifies the speed of the motor.\nmax_power [0 - 100] - specifies the maximum power of the motor when regulating the speed\nacceleration [0 - 10000] - specifies the time in msec for the motor to achieve specified speed\ndeceleration [0 - 10000] - specifies the time in msec for the motor to come to stop from full speed\nstall - specifies if stall detection is enabled or disabled (True/False)\n\n"
    ),
    _(
        "port.X.motor.run_for_degrees(degrees, speed, max_power, stop, acceleration, deceleration, stall) \nKeyword arguments:\n\ndegrees [-MAXINT - +MAXINT] - specifies the degrees to turn from current target position\nspeed [-100 - 100] - specifies the speed of the motor.\nmax_power [0 - 100] - specifies the maximum power of the motor when regulating the speed\nstop - specifies the stop state of the motor. [0 = MOTOR_STOP_FLOAT, 1 = MOTOR_STOP_BRAKE, 2 = MOTOR_STOP_HOLD]\nacceleration [0 - 10000] - specifies the time in msec for the motor to achieve specified speed\ndeceleration [0 - 10000] - specifies the time in msec for the motor to come to stop from full speed\nstall - specifies if stall detection is enabled or disabled (True/False)\n\n"
    ),
    _(
        "port.X.motor.run_for_time(msec, speed, max_power, stop, acceleration, deceleration, stall) \n\nmsec [0 - 65535] - Force the motor driver to run for msec milliseconds at speed. When the msec time expires the motor driver is forced to the stop state. If msec is negative a time of 0 (zero) milliseconds is assumed.\nmax_power [0 - 100] - specifies the maximum power of the motor when regulating the speed\nstop - specifies the stop state of the motor. [0 = MOTOR_STOP_FLOAT, 1 = MOTOR_STOP_BRAKE, 2 = MOTOR_STOP_HOLD]\nacceleration [0 - 10000] - specifies the time in msec for the motor to achieve specified speed\ndeceleration [0 - 10000] - specifies the time in msec for the motor to come to stop from full speed\nstall - specifies if stall detection is enabled or disabled (True/False)\n\n"
    ),
    _(
        "port.X.motor.run_to_position(position, speed, max_power, stop, acceleration, deceleration, stall) \nKeyword arguments:\n\nposition [-MAXINT - +MAXINT] - specifies the position of the motor.\nspeed [-100 - 100] - specifies the speed of the motor.\nmax_power [0 - 100] - specifies the maximum power of the motor when regulating the speed\nstop - specifies the stop state of the motor. [0 = MOTOR_STOP_FLOAT, 1 = MOTOR_STOP_BRAKE, 2 = MOTOR_STOP_HOLD]\nacceleration [0 - 10000] - specifies the time in msec for the motor to achieve specified speed\ndeceleration [0 - 10000] - specifies the time in msec for the motor to come to stop from full speed\nstall - specifies if stall detection is enabled or disabled (True/False)\n\n"
    ),
    _(
        "port.X.pwm(value) \nKeyword arguments:\n\nvalue [-100 - 100] Applies a PWM signal to the power pins of the port.\n\nUseful for controlling directly. The polarity of the PWM signal matches the\nsign of the value.\nA value of zero stops the PWM signal and leaves the port driver in\nthe floating state.\n"
    ),
]



"""
Contains the main Window definition for Mu's UI.

Copyright (c) 2015-2017 Nicholas H.Tollervey and others (see the AUTHORS file).

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import sys
import logging
import os.path
from PyQt5.QtCore import QSize, Qt, pyqtSignal, QTimer
from PyQt5.QtWidgets import (
    QToolBar,
    QAction,
    QDesktopWidget,
    QWidget,
    QVBoxLayout,
    QTabWidget,
    QFileDialog,
    QMessageBox,
    QLabel,
    QMainWindow,
    QStatusBar,
    QDockWidget,
    QShortcut,
    QApplication,
    QTabBar,
    QPushButton,
    QHBoxLayout,
)
from PyQt5.QtGui import QKeySequence, QStandardItemModel, QCursor
from mu import __version__
from mu.interface.dialogs import (
    ModeSelector,
    AdminDialog,
    FindReplaceDialog,
    PackageDialog,
)
from mu.interface.themes import (
    DayTheme,
    NightTheme,
    ContrastTheme,
    DEFAULT_FONT_SIZE,
)
from mu.interface.panes import (
    DebugInspector,
    DebugInspectorItem,
    PythonProcessPane,
    JupyterREPLPane,
    MicroPythonREPLPane,
    FileSystemPane,
    PlotterPane,
)
from mu.interface.editor import EditorPane
from mu.interface.widgets import DeviceSelector
from mu.resources import load_icon, load_pixmap


logger = logging.getLogger(__name__)


class ButtonBar(QToolBar):
    """
    Represents the bar of buttons across the top of the editor and defines
    their behaviour.
    """

    def __init__(self, parent):
        super().__init__(parent)
        self.setMovable(False)
        self.setIconSize(QSize(64, 64))
        self.setToolButtonStyle(3)
        self.setContextMenuPolicy(Qt.PreventContextMenu)
        self.setObjectName("StandardToolBar")
        self.reset()

    def reset(self):
        """
        Resets the button states.
        """
        self.slots = {}
        self.clear()

    def change_mode(self, mode):
        self.reset()
        self.addAction(
            name="modes",
            display_name=_("Mode"),
            tool_text=_("Change Mu's mode of behaviour."),
        )
        self.addSeparator()
        self.addAction(
            name="new",
            display_name=_("New"),
            tool_text=_("Create a new Python script."),
        )
        self.addAction(
            name="load",
            display_name=_("Load"),
            tool_text=_("Load a Python script."),
        )
        self.addAction(
            name="save",
            display_name=_("Save"),
            tool_text=_("Save the current Python script."),
        )
        self.addSeparator()

        for action in mode.actions():
            self.addAction(
                name=action["name"],
                display_name=action["display_name"],
                tool_text=action["description"],
            )

        self.addSeparator()
        self.addAction(
            name="zoom-in",
            display_name=_("Zoom-in"),
            tool_text=_("Zoom in (to make the text bigger)."),
        )
        self.addAction(
            name="zoom-out",
            display_name=_("Zoom-out"),
            tool_text=_("Zoom out (to make the text smaller)."),
        )
        self.addAction(
            name="theme",
            display_name=_("Theme"),
            tool_text=_(
                "Toggle theme between day, night or " "high contrast."
            ),
        )
        self.addSeparator()
        self.addAction(
            name="check",
            display_name=_("Check"),
            tool_text=_("Check your code for mistakes."),
        )
        if sys.version_info[:2] >= (3, 6):
            self.addAction(
                name="tidy",
                display_name=_("Tidy"),
                tool_text=_("Tidy up the layout of your code."),
            )
        self.addAction(
            name="help",
            display_name=_("Help"),
            tool_text=_("Show help about Mu in a browser."),
        )
        self.addSeparator()
        self.addAction(
            name="quit", display_name=_("Quit"), tool_text=_("Quit Mu.")
        )

    def set_responsive_mode(self, width, height):
        """
        Compact button bar for when window is very small.
        """
        font_size = min(DEFAULT_FONT_SIZE, width // 80)
        icon_size = min(64, width // 24)
        self.setIconSize(QSize(icon_size, icon_size))
        stylesheet = "QWidget{font-size: " + str(font_size) + "px;}"
        self.setStyleSheet(stylesheet)

    def addAction(self, name, display_name, tool_text):
        """
        Creates an action associated with an icon and name and adds it to the
        widget's slots.
        """
        action = QAction(
            load_icon(name), display_name, self, toolTip=tool_text
        )
        super().addAction(action)
        self.slots[name] = action

    def connect(self, name, handler, shortcut=None):
        """
        Connects a named slot to a handler function and optional hot-key
        shortcuts.
        """
        self.slots[name].pyqtConfigure(triggered=handler)
        if shortcut:
            self.slots[name].setShortcut(QKeySequence(shortcut))


class FileTabs(QTabWidget):
    """
    Extend the base class so we can override the removeTab behaviour.
    """

    def __init__(self):
        super(FileTabs, self).__init__()
        # We are implementing closable tabs manually
        self.setTabsClosable(False)
        self.setMovable(True)
        self.currentChanged.connect(self.change_tab)
        # Shrink icons to avoid giant tabs
        self.setIconSize(self.iconSize() * 0.75)

    def removeTab(self, tab_id):
        """
        Ask the user before closing the file.
        """
        window = self.nativeParentWidget()
        modified = self.widget(tab_id).isModified()
        if modified:
            msg = (
                "There is un-saved work, closing the tab will cause you "
                "to lose it."
            )
            if window.show_confirmation(msg) == QMessageBox.Cancel:
                return
        super(FileTabs, self).removeTab(tab_id)

    def addTab(self, widget, title):
        """
        Add a new tab to the switcher
        """
        # Proxy up to the real addTab
        tab_id = super(FileTabs, self).addTab(widget, title)
        # A widget to put the indicator and close button in
        container = QWidget()
        box = QHBoxLayout(container)
        # We don't want any margins on the layout, that would expand the tab
        box.setContentsMargins(0, 0, 0, 0)
        # Ensure some space between image and button
        box.setSpacing(6)
        # Counterintuitively QImage doesn't show an image, QLabel does
        state_lbl = QLabel(container)
        box.addWidget(state_lbl)
        state_lbl.setPixmap(load_pixmap("document", size=self.iconSize()))

        # Watch for status change to update the dirty indicator
        # We watch here as it's far easier to keep track of state_lbl
        # It does mean we assume all tabs are EditorPane
        @widget.modificationChanged.connect
        def on_modified():
            # Wrapped in a try/except because sometimes state_lbl is destroyed
            # by Qt before widget is GC'd by Python. Since there's no way to
            # easily check the status of state_lbl, wrapping in a try/except
            # for a RuntimeError is the simplest possible solution.
            try:
                if widget.isModified():
                    state_lbl.setPixmap(
                        load_pixmap("document-dirty", size=self.iconSize())
                    )
                else:
                    # This icon is actually empty
                    state_lbl.setPixmap(
                        load_pixmap("document", size=self.iconSize())
                    )
            except RuntimeError:  # pragma: no cover
                pass

        # Setup our own close button since we are overriding the built in one
        close_btn = QPushButton(container)
        box.addWidget(close_btn)
        close_btn.setToolTip(_("Close file"))
        close_btn.setFlat(True)
        close_btn.setIconSize(self.iconSize())
        close_btn.setIcon(load_icon("close-tab"))
        close_btn.show()

        # Handle 'clicked' events
        @close_btn.clicked.connect
        def close():
            # The tab_id isn't constant and may have changed, lookup the
            # current id of the EditorPane
            tab_id = self.indexOf(widget)
            # Close the tab
            self.removeTab(tab_id)

        container.setLayout(box)
        # Add the box, clearly it isn't a button but QTabBar actually takes
        # any QWidget not just buttons
        self.tabBar().setTabButton(tab_id, QTabBar.RightSide, container)

        # Return the index of the new page just like the reall addTab
        return tab_id

    def change_tab(self, tab_id):
        """
        Update the application title to reflect the name of the file in the
        currently selected tab.
        """
        current_tab = self.widget(tab_id)
        window = self.nativeParentWidget()
        if current_tab:
            window.update_title(current_tab.title)
        else:
            window.update_title(None)


class Window(QMainWindow):
    """
    Defines the look and characteristics of the application's main window.
    """

    title = _("Mu {}").format(__version__)
    icon = "icon"
    timer = None
    usb_checker = None
    repl = None
    plotter = None
    zooms = ("xs", "s", "m", "l", "xl", "xxl", "xxxl")  # levels of zoom.
    zoom_position = 2  # current level of zoom (as position in zooms tuple).

    _zoom_in = pyqtSignal(str)
    _zoom_out = pyqtSignal(str)
    data_received = pyqtSignal(bytes)
    open_file = pyqtSignal(str)
    load_theme = pyqtSignal(str)
    previous_folder = None
    debug_widths = None

    def __init__(self, parent=None):
        super().__init__(parent)
        # Record pane area to allow reopening where user put it in a session
        self._debugger_area = 0
        self._inspector_area = 0
        self._plotter_area = 0
        self._repl_area = 0
        self._runner_area = 0

    def wheelEvent(self, event):
        """
        Trap a CTRL-scroll event so the user is able to zoom in and out.
        """
        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.ControlModifier:
            zoom = event.angleDelta().y() > 0
            if zoom:
                self.zoom_in()
            else:
                self.zoom_out()
            event.ignore()

    def set_zoom(self):
        """
        Sets the zoom to current zoom_position level.
        """
        self._zoom_in.emit(self.zooms[self.zoom_position])

    def zoom_in(self):
        """
        Handles zooming in.
        """
        self.zoom_position = min(self.zoom_position + 1, len(self.zooms) - 1)
        self._zoom_in.emit(self.zooms[self.zoom_position])

    def zoom_out(self):
        """
        Handles zooming out.
        """
        self.zoom_position = max(self.zoom_position - 1, 0)
        self._zoom_out.emit(self.zooms[self.zoom_position])

    def connect_zoom(self, widget):
        """
        Connects a referenced widget to the zoom related signals and sets
        the zoom of the widget to the current zoom level.
        """
        self._zoom_in.connect(widget.set_zoom)
        self._zoom_out.connect(widget.set_zoom)
        widget.set_zoom(self.zooms[self.zoom_position])

    @property
    def current_tab(self):
        """
        Returns the currently focussed tab.
        """
        return self.tabs.currentWidget()

    def set_read_only(self, is_readonly):
        """
        Set all tabs read-only.
        """
        self.read_only_tabs = is_readonly
        for tab in self.widgets:
            tab.setReadOnly(is_readonly)

    def get_load_path(self, folder, extensions="*", allow_previous=True):
        """
        Displays a dialog for selecting a file to load. Returns the selected
        path. Defaults to start in the referenced folder unless a previous
        folder has been used and the allow_previous flag is True (the default
        behaviour)
        """
        if allow_previous:
            open_in = (
                folder
                if self.previous_folder is None
                else self.previous_folder
            )
        else:
            open_in = folder
        path, _ = QFileDialog.getOpenFileName(
            self.widget, "Open file", open_in, extensions
        )
        logger.debug("Getting load path: {}".format(path))
        if allow_previous:
            self.previous_folder = os.path.dirname(path)
        return path

    def get_save_path(self, folder):
        """
        Displays a dialog for selecting a file to save. Returns the selected
        path. Defaults to start in the referenced folder.
        """
        path, _ = QFileDialog.getSaveFileName(
            self.widget,
            "Save file",
            folder if self.previous_folder is None else self.previous_folder,
            "Python (*.py);;Other (*.*)",
            "Python (*.py)",
        )
        self.previous_folder = os.path.dirname(path)
        # Ensure there's a .py extension if none is provided by the user.
        # See issue #1571.
        name, ext = os.path.splitext(os.path.basename(path))
        if name and (not name.startswith(".")) and (not ext):
            # The file is not a . (dot) file and there's no extension, so add
            # .py as default.
            path += ".py"
        logger.debug("Getting save path: {}".format(path))
        return path

    def get_microbit_path(self, folder):
        """
        Displays a dialog for locating the location of the BBC micro:bit in the
        host computer's filesystem. Returns the selected path. Defaults to
        start in the referenced folder.
        """
        path = QFileDialog.getExistingDirectory(
            self.widget,
            "Locate BBC micro:bit",
            folder if self.previous_folder is None else self.previous_folder,
            QFileDialog.ShowDirsOnly,
        )
        self.previous_folder = os.path.dirname(path)
        logger.debug("Getting micro:bit path: {}".format(path))
        return path

    def add_tab(self, path, text, api, newline):
        """
        Adds a tab with the referenced path and text to the editor.
        """
        new_tab = EditorPane(path, text, newline)
        new_tab.connect_margin(self.breakpoint_toggle)
        new_tab_index = self.tabs.addTab(new_tab, new_tab.label)
        new_tab.set_api(api)

        @new_tab.modificationChanged.connect
        def on_modified():
            modified_tab_index = self.tabs.indexOf(new_tab)
            # Update tab label & window title
            # Tab dirty indicator is managed in FileTabs.addTab
            self.tabs.setTabText(modified_tab_index, new_tab.label)
            self.update_title(new_tab.title)

        @new_tab.open_file.connect
        def on_open_file(file):
            # Bubble the signal up
            self.open_file.emit(file)

        new_tab.context_menu.connect(self.on_context_menu)

        self.tabs.setCurrentIndex(new_tab_index)
        self.connect_zoom(new_tab)
        self.set_theme(self.theme)
        new_tab.setFocus()
        if self.read_only_tabs:
            new_tab.setReadOnly(self.read_only_tabs)
        return new_tab

    def focus_tab(self, tab):
        """
        Force focus on the referenced tab.
        """
        index = self.tabs.indexOf(tab)
        self.tabs.setCurrentIndex(index)
        tab.setFocus()

    @property
    def tab_count(self):
        """
        Returns the number of active tabs.
        """
        return self.tabs.count()

    @property
    def widgets(self):
        """
        Returns a list of references to the widgets representing tabs in the
        editor.
        """
        return [self.tabs.widget(i) for i in range(self.tab_count)]

    @property
    def modified(self):
        """
        Returns a boolean indication if there are any modified tabs in the
        editor.
        """
        for widget in self.widgets:
            if widget.isModified():
                return True
        return False

    def on_context_menu(self):
        """
        Called when a user right-clicks on an editor pane.

        If the REPL is active AND there is selected text in the current editor
        pane, modify the default context menu to include a paste to REPL
        option. Otherwise, just display the default context menu.
        """
        # Create a standard context menu.
        menu = self.current_tab.createStandardContextMenu()
        # Flag to indicate if there is a section of text highlighted.
        has_selection = any([x > -1 for x in self.current_tab.getSelection()])
        # Flag to indicate if the REPL pane is active.
        has_repl = hasattr(self, "repl") and self.repl
        # Flag to indicate if the Python process runner pane is active and in
        # interactive Python3 mode (pasting into other "standard" Python modes
        # doesn't make sense).
        has_runner = (
            hasattr(self, "process_runner")
            and self.process_runner
            and self.process_runner.is_interactive
        )
        if has_selection and (has_repl or has_runner):
            # Text selected with REPL/process context, so add the bespoke
            # "copy to repl" item to the standard context menu to handle this
            # situation.
            actions = menu.actions()
            copy_to_repl = QAction(_("Copy selected text to REPL"), menu)
            copy_to_repl.triggered.connect(self.copy_to_repl)
            menu.insertAction(actions[0], copy_to_repl)
            menu.insertSeparator(actions[0])
        # Display menu.
        menu.exec_(QCursor.pos())

    def copy_to_repl(self):
        """
        Copies currently selected text in the editor, into the active REPL
        widget and sets focus to the REPL widget. The final line pasted into
        the REPL waits for RETURN to be pressed by the user (this appears to be
        the default behaviour for pasting into the REPL widget).
        """
        line_from, pos_from, line_to, pos_to = self.current_tab.getSelection()
        lines = self.current_tab.text().split("\n")
        if line_from == line_to:
            # Paste a fragment from an individual line.
            to_paste = lines[line_from][pos_from:pos_to]
        else:
            # Multi-line paste (paste all lines selected).
            selected = lines[line_from : line_to + 1]
            # Ensure the correct indentation.
            # If the first line starts with whitespace, work out the number of
            # spaces and deduct said number of whitespace from start of all
            # other lines. That's perhaps the best we can do.
            if selected and selected[0].startswith(" "):
                indent = 0
                for char in selected[0]:
                    if char.isspace():
                        indent += 1
                    else:
                        break
                selected = [line[indent:] for line in selected]
            to_paste = "\n".join(selected)
        logger.info("Pasting to REPL")
        logger.info("\n" + to_paste)
        clipboard = QApplication.clipboard()
        clipboard.setText(to_paste)
        if hasattr(self, "repl_pane") and self.repl_pane:
            self.repl_pane.paste()
            self.repl_pane.setFocus()
        elif hasattr(self, "process_runner") and self.process_runner:
            self.process_runner.paste()
            self.process_runner.setFocus()

    def on_stdout_write(self, data):
        """
        Called when either a running script or the REPL write to STDOUT.
        """
        self.data_received.emit(data)

    def add_filesystem(self, home, file_manager, board_name="board"):
        """
        Adds the file system pane to the application.
        """
        self.fs_pane = FileSystemPane(home)

        @self.fs_pane.open_file.connect
        def on_open_file(file):
            # Bubble the signal up
            self.open_file.emit(file)

        self.fs = QDockWidget(_("Filesystem on ") + board_name)
        self.fs.setWidget(self.fs_pane)
        self.fs.setFeatures(QDockWidget.DockWidgetMovable)
        self.fs.setAllowedAreas(Qt.BottomDockWidgetArea)
        self.addDockWidget(Qt.BottomDockWidgetArea, self.fs)
        self.fs_pane.setFocus()
        file_manager.on_list_files.connect(self.fs_pane.on_ls)
        self.fs_pane.list_files.connect(file_manager.ls)
        self.fs_pane.microbit_fs.put.connect(file_manager.put)
        self.fs_pane.microbit_fs.delete.connect(file_manager.delete)
        self.fs_pane.microbit_fs.list_files.connect(file_manager.ls)
        self.fs_pane.local_fs.get.connect(file_manager.get)
        self.fs_pane.local_fs.put.connect(file_manager.put)
        self.fs_pane.local_fs.list_files.connect(file_manager.ls)
        file_manager.on_put_file.connect(self.fs_pane.microbit_fs.on_put)
        file_manager.on_delete_file.connect(self.fs_pane.microbit_fs.on_delete)
        file_manager.on_get_file.connect(self.fs_pane.local_fs.on_get)
        file_manager.on_list_fail.connect(self.fs_pane.on_ls_fail)
        file_manager.on_put_fail.connect(self.fs_pane.on_put_fail)
        file_manager.on_delete_fail.connect(self.fs_pane.on_delete_fail)
        file_manager.on_get_fail.connect(self.fs_pane.on_get_fail)
        self.connect_zoom(self.fs_pane)
        return self.fs_pane

    def add_micropython_repl(self, name, connection):
        """
        Adds a MicroPython based REPL pane to the application.
        """
        repl_pane = MicroPythonREPLPane(connection)
        connection.data_received.connect(repl_pane.process_tty_data)
        self.add_repl(repl_pane, name)

    def add_micropython_plotter(self, name, connection, data_flood_handler):
        """
        Adds a plotter that reads data from a serial connection.
        """
        plotter_pane = PlotterPane()
        connection.data_received.connect(plotter_pane.process_tty_data)
        plotter_pane.data_flood.connect(data_flood_handler)
        self.add_plotter(plotter_pane, name)

    def add_python3_plotter(self, mode):
        """
        Add a plotter that reads from either the REPL or a running script.
        Since this function will only be called when either the REPL or a
        running script are running (but not at the same time), it'll just grab
        data emitted by the REPL or script via data_received.
        """
        plotter_pane = PlotterPane()
        self.data_received.connect(plotter_pane.process_tty_data)
        plotter_pane.data_flood.connect(mode.on_data_flood)
        self.add_plotter(plotter_pane, _("Python3 data tuple"))

    def add_jupyter_repl(self, kernel_manager, kernel_client):
        """
        Adds a Jupyter based REPL pane to the application.
        """
        kernel_manager.kernel.gui = "qt4"
        kernel_client.start_channels()
        ipython_widget = JupyterREPLPane()
        ipython_widget.kernel_manager = kernel_manager
        ipython_widget.kernel_client = kernel_client
        ipython_widget.on_append_text.connect(self.on_stdout_write)
        self.add_repl(ipython_widget, _("Python3 (Jupyter)"))

    def add_repl(self, repl_pane, name):
        """
        Adds the referenced REPL pane to the application.
        """
        self.repl_pane = repl_pane
        self.repl = QDockWidget(_("{} REPL").format(name))
        self.repl.setWidget(repl_pane)
        self.repl.setFeatures(QDockWidget.DockWidgetMovable)
        self.repl.setAllowedAreas(
            Qt.BottomDockWidgetArea
            | Qt.LeftDockWidgetArea
            | Qt.RightDockWidgetArea
        )
        area = self._repl_area or Qt.BottomDockWidgetArea
        self.addDockWidget(area, self.repl)
        self.connect_zoom(self.repl_pane)
        self.repl_pane.set_theme(self.theme)
        self.repl_pane.setFocus()

    def add_plotter(self, plotter_pane, name):
        """
        Adds the referenced plotter pane to the application.
        """
        self.plotter_pane = plotter_pane
        self.plotter = QDockWidget(_("{} Plotter").format(name))
        self.plotter.setWidget(plotter_pane)
        self.plotter.setFeatures(QDockWidget.DockWidgetMovable)
        self.plotter.setAllowedAreas(
            Qt.BottomDockWidgetArea
            | Qt.LeftDockWidgetArea
            | Qt.RightDockWidgetArea
        )
        area = self._plotter_area or Qt.BottomDockWidgetArea
        self.addDockWidget(area, self.plotter)
        self.plotter_pane.set_theme(self.theme)
        self.plotter_pane.setFocus()

    def add_python3_runner(
        self,
        interpreter,
        script_name,
        working_directory,
        interactive=False,
        debugger=False,
        command_args=None,
        envars=None,
        python_args=None,
    ):
        """
        Display console output for the interpreter with the referenced
        pythonpath running the referenced script.

        The script will be run within the workspace_path directory.

        If interactive is True (default is False) the Python process will
        run in interactive mode (dropping the user into the REPL when the
        script completes).

        If debugger is True (default is False) the script will be run within
        a debug runner session. The debugger overrides the interactive flag
        (you cannot run the debugger in interactive mode).

        If there is a list of command_args (the default is None) then these
        will be passed as further arguments into the command run in the
        new process.

        If envars is given, these will become part of the environment context
        of the new chlid process.

        If python_args is given, these will be passed as arguments to the
        Python runtime used to launch the child process.
        """
        self.process_runner = PythonProcessPane(self)
        self.runner = QDockWidget(
            _("Running: {}").format(os.path.basename(script_name))
        )
        self.runner.setWidget(self.process_runner)
        self.runner.setFeatures(QDockWidget.DockWidgetMovable)
        self.runner.setAllowedAreas(
            Qt.BottomDockWidgetArea
            | Qt.LeftDockWidgetArea
            | Qt.RightDockWidgetArea
        )
        self.process_runner.debugger = debugger
        if debugger:
            area = self._debugger_area or Qt.BottomDockWidgetArea
        else:
            area = self._runner_area or Qt.BottomDockWidgetArea
        self.addDockWidget(area, self.runner)
        logger.info(
            "About to start_process: %r, %r, %r, %r, %r, %r, %r, %r",
            interpreter,
            script_name,
            working_directory,
            interactive,
            debugger,
            command_args,
            envars,
            python_args,
        )

        self.process_runner.start_process(
            interpreter,
            script_name,
            working_directory,
            interactive,
            debugger,
            command_args,
            envars,
            python_args,
        )
        self.process_runner.setFocus()
        self.process_runner.on_append_text.connect(self.on_stdout_write)
        self.connect_zoom(self.process_runner)
        return self.process_runner

    def add_debug_inspector(self):
        """
        Display a debug inspector to view the call stack.
        """
        self.debug_inspector = DebugInspector()
        self.debug_model = QStandardItemModel()
        self.debug_inspector.setModel(self.debug_model)
        self.inspector = QDockWidget(_("Debug Inspector"))
        self.inspector.setWidget(self.debug_inspector)
        self.inspector.setFeatures(QDockWidget.DockWidgetMovable)
        self.inspector.setAllowedAreas(
            Qt.BottomDockWidgetArea
            | Qt.LeftDockWidgetArea
            | Qt.RightDockWidgetArea
        )
        area = self._inspector_area or Qt.RightDockWidgetArea
        self.addDockWidget(area, self.inspector)
        self.connect_zoom(self.debug_inspector)
        # Setup the inspector headers and restore column widths
        self.debug_model.setHorizontalHeaderLabels([_("Name"), _("Value")])
        if self.debug_widths:
            for col, width in enumerate(self.debug_widths):
                self.debug_inspector.setColumnWidth(col, width)

    def update_debug_inspector(self, locals_dict):
        """
        Given the contents of a dict representation of the locals in the
        current stack frame, update the debug inspector with the new values.
        """
        excluded_names = ["__builtins__", "__debug_code__", "__debug_script__"]
        names = sorted([x for x in locals_dict if x not in excluded_names])

        # Remove rows so we keep the same column layouts if manually set
        while self.debug_model.rowCount() > 0:
            self.debug_model.removeRow(0)
        for name in names:
            item_to_expand = None
            try:
                # DANGER!
                val = eval(locals_dict[name])
            except Exception:
                val = None
            if isinstance(val, list):
                # Show a list consisting of rows of position/value
                list_item = DebugInspectorItem(name)
                item_to_expand = list_item
                for i, i_val in enumerate(val):
                    list_item.appendRow(
                        [
                            DebugInspectorItem(str(i)),
                            DebugInspectorItem(repr(i_val)),
                        ]
                    )
                self.debug_model.appendRow(
                    [
                        list_item,
                        DebugInspectorItem(
                            _("(A list of {} items.)").format(len(val))
                        ),
                    ]
                )
            elif isinstance(val, dict):
                # Show a dict consisting of rows of key/value pairs.
                dict_item = DebugInspectorItem(name)
                item_to_expand = dict_item
                for k, k_val in val.items():
                    dict_item.appendRow(
                        [
                            DebugInspectorItem(repr(k)),
                            DebugInspectorItem(repr(k_val)),
                        ]
                    )
                self.debug_model.appendRow(
                    [
                        dict_item,
                        DebugInspectorItem(
                            _("(A dict of {} items.)").format(len(val))
                        ),
                    ]
                )
            else:
                self.debug_model.appendRow(
                    [
                        DebugInspectorItem(name),
                        DebugInspectorItem(locals_dict[name]),
                    ]
                )
            # Expand dicts/list with names matching old expanded entries
            if (
                hasattr(self, "debug_inspector")
                and name in self.debug_inspector.expanded_dicts
                and item_to_expand is not None
            ):
                self.debug_inspector.expand(
                    self.debug_model.indexFromItem(item_to_expand)
                )

    def remove_filesystem(self):
        """
        Removes the file system pane from the application.
        """
        if hasattr(self, "fs") and self.fs:
            self.fs_pane = None
            self.fs.setParent(None)
            self.fs.deleteLater()
            self.fs = None

    def remove_repl(self):
        """
        Removes the REPL pane from the application.
        """
        if self.repl:
            self._repl_area = self.dockWidgetArea(self.repl)
            self.repl_pane = None
            self.repl.setParent(None)
            self.repl.deleteLater()
            self.repl = None

    def remove_plotter(self):
        """
        Removes the plotter pane from the application.
        """
        if self.plotter:
            self._plotter_area = self.dockWidgetArea(self.plotter)
            self.plotter_pane = None
            self.plotter.setParent(None)
            self.plotter.deleteLater()
            self.plotter = None

    def remove_python_runner(self):
        """
        Removes the runner pane from the application.
        """
        if hasattr(self, "runner") and self.runner:
            if self.process_runner.debugger:
                self._debugger_area = self.dockWidgetArea(self.runner)
            else:
                self._runner_area = self.dockWidgetArea(self.runner)
            self.process_runner = None
            self.runner.setParent(None)
            self.runner.deleteLater()
            self.runner = None

    def remove_debug_inspector(self):
        """
        Removes the debug inspector pane from the application.
        """
        if hasattr(self, "inspector") and self.inspector:
            width = self.debug_inspector.columnWidth
            self.debug_widths = width(0), width(1)
            self._inspector_area = self.dockWidgetArea(self.inspector)
            self.debug_inspector = None
            self.debug_model = None
            self.inspector.setParent(None)
            self.inspector.deleteLater()
            self.inspector = None

    def set_theme(self, theme):
        """
        Sets the theme for the REPL and editor tabs.
        """
        self.theme = theme
        self.load_theme.emit(theme)
        if theme == "contrast":
            new_theme = ContrastTheme
            new_icon = "theme_day"
        elif theme == "night":
            new_theme = NightTheme
            new_icon = "theme_contrast"
        else:
            new_theme = DayTheme
            new_icon = "theme"
        for widget in self.widgets:
            widget.set_theme(new_theme)
        self.button_bar.slots["theme"].setIcon(load_icon(new_icon))
        if hasattr(self, "repl") and self.repl:
            self.repl_pane.set_theme(theme)
        if hasattr(self, "plotter") and self.plotter:
            self.plotter_pane.set_theme(theme)
        if hasattr(self, "runner") and self.runner:
            self.process_runner.set_theme(theme)

    def set_checker_icon(self, icon):
        """
        Set the status icon to use on the check button
        """
        self.button_bar.slots["check"].setIcon(load_icon(icon))
        timer = QTimer()

        @timer.timeout.connect
        def reset():
            self.button_bar.slots["check"].setIcon(load_icon("check"))
            timer.stop()

        timer.start(500)

    def show_admin(self, log, settings, packages, mode, device_list):
        """
        Display the administrative dialog with referenced content of the log
        and settings. Return a dictionary of the settings that may have been
        changed by the admin dialog.
        """
        admin_box = AdminDialog(self)
        admin_box.setup(log, settings, packages, mode, device_list)
        result = admin_box.exec()
        if result:
            return admin_box.settings()
        else:
            return {}

    def sync_packages(self, to_remove, to_add):
        """
        Display a modal dialog that indicates the status of the add/remove
        package management operation.
        """
        package_box = PackageDialog(self)
        package_box.setup(to_remove, to_add)
        package_box.exec()

    def show_message(self, message, information=None, icon=None):
        """
        Displays a modal message to the user.

        If information is passed in this will be set as the additional
        informative text in the modal dialog.

        Since this mechanism will be used mainly for warning users that
        something is awry the default icon is set to "Warning". It's possible
        to override the icon to one of the following settings: NoIcon,
        Question, Information, Warning or Critical.
        """
        message_box = QMessageBox(self)
        message_box.setText(message)
        message_box.setWindowTitle("Mu")
        if information:
            message_box.setInformativeText(information)
        if icon and hasattr(message_box, icon):
            message_box.setIcon(getattr(message_box, icon))
        else:
            message_box.setIcon(message_box.Warning)
        logger.debug(message)
        logger.debug(information)
        message_box.exec()

    def show_confirmation(self, message, information=None, icon=None):
        """
        Displays a modal message to the user to which they need to confirm or
        cancel.

        If information is passed in this will be set as the additional
        informative text in the modal dialog.

        Since this mechanism will be used mainly for warning users that
        something is awry the default icon is set to "Warning". It's possible
        to override the icon to one of the following settings: NoIcon,
        Question, Information, Warning or Critical.
        """
        message_box = QMessageBox(self)
        message_box.setText(message)
        message_box.setWindowTitle(_("Mu"))
        if information:
            message_box.setInformativeText(information)
        if icon and hasattr(message_box, icon):
            message_box.setIcon(getattr(message_box, icon))
        else:
            message_box.setIcon(message_box.Warning)
        message_box.setStandardButtons(message_box.Cancel | message_box.Ok)
        message_box.setDefaultButton(message_box.Cancel)
        logger.debug(message)
        logger.debug(information)
        return message_box.exec()

    def update_title(self, filename=None):
        """
        Updates the title bar of the application. If a filename (representing
        the name of the file currently the focus of the editor) is supplied,
        append it to the end of the title.
        """
        title = self.title
        if filename:
            title += " - " + filename
        self.setWindowTitle(title)

    def screen_size(self):
        """
        Returns an (width, height) tuple with the screen geometry.
        """
        screen = QDesktopWidget().screenGeometry()
        return screen.width(), screen.height()

    def size_window(self, x=None, y=None, w=None, h=None):
        """
        Makes the editor 80% of the width*height of the screen and centres it
        when none of x, y, w and h is passed in; otherwise uses the passed in
        values to position and size the editor window.

        If the X or Y value will be off the screen, these are reset to None
        (thus stopping the window being drawn in a hard-to-reach place). See
        issue #1613 for context.
        """
        screen_width, screen_height = self.screen_size()
        w = int(screen_width * 0.8) if w is None else w
        h = int(screen_height * 0.8) if h is None else h
        self.resize(w, h)
        size = self.geometry()
        # Ensure the window isn't added off the screen.
        if x and (x <= 0 or x > screen_width):
            x = None
        if y and (y <= 0 or y > screen_height):
            y = None
        x = (screen_width - size.width()) // 2 if x is None else x
        y = (screen_height - size.height()) // 2 if y is None else y
        self.move(x, y)

    def reset_annotations(self):
        """
        Resets the state of annotations on the current tab.
        """
        self.current_tab.reset_annotations()

    def annotate_code(self, feedback, annotation_type):
        """
        Given a list of annotations about the code in the current tab, add
        the annotations to the editor window so the user can make appropriate
        changes.
        """
        self.current_tab.annotate_code(feedback, annotation_type)

    def show_annotations(self):
        """
        Show the annotations added to the current tab.
        """
        self.current_tab.show_annotations()

    def setup(self, breakpoint_toggle, theme):
        """
        Sets up the window.

        Defines the various attributes of the window and defines how the user
        interface is laid out.
        """
        self.theme = theme
        self.breakpoint_toggle = breakpoint_toggle
        # Give the window a default icon, title and minimum size.
        self.setWindowIcon(load_icon(self.icon))
        self.update_title()
        self.read_only_tabs = False
        screen_width, screen_height = self.screen_size()
        self.setMinimumSize(screen_width // 2, screen_height // 2)
        self.setTabPosition(Qt.AllDockWidgetAreas, QTabWidget.North)
        self.widget = QWidget()
        widget_layout = QVBoxLayout()
        self.widget.setLayout(widget_layout)
        self.button_bar = ButtonBar(self.widget)
        self.tabs = FileTabs()
        self.setCentralWidget(self.tabs)
        self.status_bar = StatusBar(parent=self)
        self.setStatusBar(self.status_bar)
        self.addToolBar(self.button_bar)
        self.show()

    def resizeEvent(self, resizeEvent):
        """
        Respond to window getting too small for the button bar to fit well.
        """
        size = resizeEvent.size()
        self.button_bar.set_responsive_mode(size.width(), size.height())

    def select_mode(self, modes, current_mode):
        """
        Display the mode selector dialog and return the result.
        """
        mode_select = ModeSelector(self)
        mode_select.setup(modes, current_mode)
        mode_select.exec()
        try:
            return mode_select.get_mode()
        except Exception:
            return None

    def change_mode(self, mode):
        """
        Given a an object representing a mode, recreates the button bar with
        the expected functionality.
        """
        self.button_bar.change_mode(mode)
        # Update the autocomplete / tooltip APIs for each tab to the new mode.
        api = mode.api()
        for widget in self.widgets:
            widget.set_api(api)

    def set_usb_checker(self, duration, callback):
        """
        Sets up a timer that polls for USB changes via the "callback" every
        "duration" seconds.
        """
        self.usb_checker = QTimer()
        self.usb_checker.timeout.connect(callback)
        self.usb_checker.start(duration * 1000)

    def set_timer(self, duration, callback):
        """
        Set a repeating timer to call "callback" every "duration" seconds.
        """
        self.timer = QTimer()
        self.timer.timeout.connect(callback)
        self.timer.start(duration * 1000)  # Measured in milliseconds.

    def stop_timer(self):
        """
        Stop the repeating timer.
        """
        if self.timer:
            self.timer.stop()
            self.timer = None

    def connect_tab_rename(self, handler, shortcut):
        """
        Connect the double-click event on a tab and the keyboard shortcut to
        the referenced handler (causing the Save As dialog).
        """
        self.tabs.shortcut = QShortcut(QKeySequence(shortcut), self)
        self.tabs.shortcut.activated.connect(handler)
        self.tabs.tabBarDoubleClicked.connect(handler)

    def open_directory_from_os(self, path):
        """
        Given the path to a directory, open the OS's built in filesystem
        explorer for that path. Works with Windows, OSX and Linux.
        """
        if sys.platform == "win32":
            # Windows
            os.startfile(path)
        elif sys.platform == "darwin":
            # OSX
            os.system('open "{}"'.format(path))
        else:
            # Assume freedesktop.org on unix-y.
            os.system('xdg-open "{}"'.format(path))

    def connect_find_replace(self, handler, shortcut):
        """
        Create a keyboard shortcut and associate it with a handler for doing
        a find and replace.
        """
        self.find_replace_shortcut = QShortcut(QKeySequence(shortcut), self)
        self.find_replace_shortcut.activated.connect(handler)

    def connect_find_again(self, handlers, shortcut):
        """
        Create keyboard shortcuts and associate them with handlers for doing
        a find again in forward or backward direction. Any given shortcut
        will be used for forward find again, while Shift+shortcut will find
        again backwards.
        """
        forward, backward = handlers
        self.find_again_shortcut = QShortcut(QKeySequence(shortcut), self)
        self.find_again_shortcut.activated.connect(forward)
        backward_shortcut = QKeySequence("Shift+" + shortcut)
        self.find_again_backward_shortcut = QShortcut(backward_shortcut, self)
        self.find_again_backward_shortcut.activated.connect(backward)

    def show_find_replace(self, find, replace, global_replace):
        """
        Display the find/replace dialog. If the dialog's OK button was clicked
        return a tuple containing the find term, replace term and global
        replace flag.
        """
        finder = FindReplaceDialog(self)
        finder.setup(find, replace, global_replace)
        if finder.exec():
            return (finder.find(), finder.replace(), finder.replace_flag())

    def replace_text(self, target_text, replace, global_replace):
        """
        Given target_text, replace the first instance after the cursor with
        "replace". If global_replace is true, replace all instances of
        "target". Returns the number of times replacement has occurred.
        """
        if not self.current_tab:
            return 0
        if global_replace:
            counter = 0
            found = self.current_tab.findFirst(
                target_text, False, True, False, False, line=0, index=0
            )
            if found:
                counter += 1
                self.current_tab.replace(replace)
                while self.current_tab.findNext():
                    self.current_tab.replace(replace)
                    counter += 1
            return counter
        else:
            found = self.current_tab.findFirst(
                target_text, False, True, False, True
            )
            if found:
                self.current_tab.replace(replace)
                return 1
            else:
                return 0

    def highlight_text(self, target_text, forward=True):
        """
        Highlight the first match from the current position of the cursor in
        the current tab for the target_text. Returns True if there's a match.
        """
        if self.current_tab:
            line = -1
            index = -1
            if not forward:
                # Workaround for `findFirst(forward=False)` not advancing
                # backwards: pass explicit line and index values.
                line, index, _el, _ei = self.current_tab.getSelection()
            return self.current_tab.findFirst(
                target_text,  # Text to find,
                False,  # Treat as regular expression
                True,  # Case sensitive search
                False,  # Whole word matches only
                True,  # Wrap search
                forward=forward,  # Forward search
                line=line,  # -1 starts at current position
                index=index,  # -1 starts at current position
            )
        else:
            return False

    def connect_toggle_comments(self, handler, shortcut):
        """
        Create a keyboard shortcut and associate it with a handler for toggling
        comments on highlighted lines.
        """
        self.toggle_comments_shortcut = QShortcut(QKeySequence(shortcut), self)
        self.toggle_comments_shortcut.activated.connect(handler)

    def toggle_comments(self):
        """
        Toggle comments on/off for all selected line in the currently active
        tab.
        """
        if self.current_tab:
            self.current_tab.toggle_comments()

    def show_device_selector(self):
        """
        Reveals the device selector in the status bar
        """
        self.status_bar.device_selector.setHidden(False)

    def hide_device_selector(self):
        """
        Hides the device selector in the status bar
        """
        self.status_bar.device_selector.setHidden(True)


class StatusBar(QStatusBar):
    """
    Defines the look and behaviour of the status bar along the bottom of the
    UI.
    """

    def __init__(self, parent=None, mode="python"):
        super().__init__(parent)
        self.mode = mode
        self.msg_duration = 5

        # Mode selector.
        self.mode_label = QLabel()
        self.mode_label.setToolTip(_("Mu's current mode of behaviour."))
        self.addPermanentWidget(self.mode_label)
        self.set_mode(mode)

        # Device selector.
        self.device_selector = DeviceSelector()
        self.device_selector.setHidden(True)
        self.addPermanentWidget(self.device_selector)

        # Logs viewer
        self.logs_label = QLabel()
        self.logs_label.setObjectName("AdministrationLabel")
        self.logs_label.setPixmap(load_pixmap("logs", size=QSize(24, 24)))
        self.logs_label.setToolTip(_("Mu Administration"))
        self.addPermanentWidget(self.logs_label)

    def connect_logs(self, handler, shortcut):
        """
        Connect the mouse press event and keyboard shortcut for the log widget
        to the referenced handler function.
        """
        self.logs_label.shortcut = QShortcut(
            QKeySequence(shortcut), self.parent()
        )
        self.logs_label.shortcut.activated.connect(handler)
        self.logs_label.mousePressEvent = handler

    def connect_mode(self, handler, shortcut):
        """
        Connect the mouse press event and keyboard shortcut for the mode widget
        to the referenced handler function.
        """
        self.mode_label.shortcut = QShortcut(
            QKeySequence(shortcut), self.parent()
        )
        self.mode_label.shortcut.activated.connect(handler)
        self.mode_label.mousePressEvent = handler

    def set_message(self, message, pause=5000):
        """
        Displays a message in the status bar for a certain period of time.
        """
        self.showMessage(message, pause)

    def set_mode(self, mode):
        """
        Updates the mode label to the new mode.
        """
        self.mode_label.setText(mode)

    def device_connected(self, device):
        """
        Show a tooltip whenever a new device connects
        """
        if device.board_name:
            msg = _("Detected new {} device: {}.").format(
                device.long_mode_name, device.board_name
            )
        else:
            msg = _("Detected new {} device.").format(device.long_mode_name)

        self.set_message(msg, self.msg_duration * 1000)



"""
The Web mode for the Mu editor, using Flask.

Copyright (c) 2015-2017 Nicholas H.Tollervey and others (see the AUTHORS file).

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import logging
import webbrowser
from mu.modes.base import BaseMode
from mu.modes.api import PYTHON3_APIS, SHARED_APIS, FLASK_APIS
from mu.resources import load_icon
from mu.logic import read_and_decode
from ..virtual_environment import venv


logger = logging.getLogger(__name__)


MUWEB_TEMPLATE = """# Wraps the user's Flask application with boilerplate.
from {} import app


app.run()
"""


CODE_TEMPLATE = """# A simple web application.
from flask import Flask, render_template


app = Flask(__name__)


@app.route("/")
def index():
    return render_template('index.html')
"""


class WebMode(BaseMode):
    """
    Represents the functionality required by the WWW mode.
    """

    name = _("Web")
    short_name = "web"
    description = _('Build simple websites with the "Flask" web framework.')
    icon = "web"
    runner = None
    save_timeout = 0  # User has to explicitly save web application.
    file_extensions = ["css", "html"]
    code_template = CODE_TEMPLATE

    def actions(self):
        """
        Return an ordered list of actions provided by this module. An action
        is a name (also used to identify the icon) , description, and handler.
        """
        return [
            {
                "name": "run",
                "display_name": _("Run"),
                "description": _("Run the web server."),
                "handler": self.run_toggle,
                "shortcut": "F5",
            },
            {
                "name": "browse",
                "display_name": _("Browse"),
                "description": _("Open your website in a browser."),
                "handler": self.browse,
                "shortcut": "Ctrl+Shift+B",
            },
            {
                "name": "templates",
                "display_name": _("Templates"),
                "description": _("Load HTML templates used by your website."),
                "handler": self.load_templates,
                "shortcut": "Ctrl+Shift+T",
            },
            {
                "name": "css",
                "display_name": _("CSS"),
                "description": _("Load CSS files used by your website."),
                "handler": self.load_css,
                "shortcut": "Ctrl+Shift+C",
            },
            {
                "name": "static",
                "display_name": _("Images"),
                "description": _(
                    "Open the directory containing images used "
                    "by your website."
                ),
                "handler": self.show_images,
                "shortcut": "Ctrl+Shift+I",
            },
        ]

    def api(self):
        """
        Return a list of API specifications to be used by auto-suggest and call
        tips.
        """
        return SHARED_APIS + PYTHON3_APIS + FLASK_APIS

    def run_toggle(self, event):
        """
        Handles the toggling of the run button to start/stop the server.
        """
        if self.runner:
            self.stop_server()
            run_slot = self.view.button_bar.slots["run"]
            run_slot.setIcon(load_icon("run"))
            run_slot.setText(_("Run"))
            run_slot.setToolTip(_("Run the web server."))
            self.set_buttons(modes=True)
        else:
            self.start_server()
            if self.runner:
                run_slot = self.view.button_bar.slots["run"]
                run_slot.setIcon(load_icon("stop"))
                run_slot.setText(_("Stop"))
                run_slot.setToolTip(_("Stop the web server."))
                self.set_buttons(modes=False)

    def start_server(self):
        """
        Run the current file as a webserver.
        """
        # Grab the Python file.
        tab = self.view.current_tab
        if tab is None:
            logger.debug("There is no active text editor.")
            self.stop_server()
            return
        if tab.path is None:
            # Unsaved file.
            self.editor.save()
        if tab.path:
            # Check it's a Python file.
            if not tab.path.lower().endswith(".py"):
                # Oops... show a helpful message and stop.
                msg = _("This is not a Python file!")
                info = _(
                    "Mu is only able to serve a Python file. Please make "
                    "sure the current tab in Mu is the one for your web "
                    "application and then try again."
                )
                self.view.show_message(msg, info)
                self.stop_server()
                return
            # If needed, save the script.
            if tab.isModified():
                self.editor.save_tab_to_file(tab)
            # Check for template files.
            template_path = os.path.join(
                os.path.dirname(os.path.abspath(tab.path)), "templates"
            )
            if not os.path.isdir(template_path):
                # Oops... show a helpful message and stop.
                msg = _("Cannot find template directory!")
                info = _(
                    "To serve your web application, there needs to be a "
                    "'templates' directory in the same place as your web "
                    "application's Python code. Please fix this and try "
                    "again. (Hint: Mu was expecting the `templates` directory "
                    "to be here: " + template_path + ")"
                )
                self.view.show_message(msg, info)
                self.stop_server()
                return
            logger.debug(tab.text())
            envars = self.editor.envars
            envars["FLASK_APP"] = os.path.basename(tab.path)
            envars["FLASK_ENV"] = "development"
            envars["LC_ALL"] = "en_GB.UTF8"
            envars["LANG"] = "en_GB.UTF8"
            args = ["-m", "flask", "run"]
            cwd = os.path.dirname(tab.path)
            self.runner = self.view.add_python3_runner(
                venv.interpreter,
                "",
                cwd,
                interactive=False,
                envars=envars,
                python_args=args,
            )
            logger.debug("Starting Flask app.")
            self.runner.process.waitForStarted()

    def stop_server(self):
        """
        Stop the currently running web server.
        """
        logger.debug("Stopping Flask app.")
        if self.runner:
            self.runner.stop_process()
            self.runner = None
        self.view.remove_python_runner()

    def stop(self):
        """
        Called by logic.quit, so if the server is running, stop it.
        """
        self.stop_server()

    def open_file(self, path):
        """
        Open the referenced file (html / css).
        """
        return read_and_decode(path)

    def load_templates(self, event):
        """
        Open the directory containing the HTML template views used by Flask.

        This should open the host OS's file system explorer so users can drag
        new files into the opened folder.
        """
        templates_dir = self.assets_dir("templates")
        logger.info(templates_dir)
        self.editor.load(default_path=templates_dir)

    def load_css(self, event):
        """
        Open the directory containing the HTML template views used by Flask.

        This should open the host OS's file system explorer so users can drag
        new files into the opened folder.
        """
        css_path = os.path.join("static", "css")
        css_dir = self.assets_dir(css_path)
        logger.info(css_dir)
        self.editor.load(default_path=css_dir)

    def show_images(self, event):
        """
        Open the directory containing the static image assets used by Flask.

        This should open the host OS's file system explorer so users can drag
        new files into the opened folder.
        """
        img_path = os.path.join("static", "img")
        img_dir = self.assets_dir(img_path)
        logger.info(img_dir)
        self.view.open_directory_from_os(img_dir)

    def browse(self, event):
        """
        Open the default browser to http://127.0.0.1:5080/ if the local web
        server is running, otherwise display a helpful message.
        """
        if self.runner:
            url = "http://127.0.0.1:5000/"
            logger.info("Opening local website at: {}".format(url))
            webbrowser.open(url)
        else:
            logger.info("Attempted to load website, but server not running.")
            msg = _("Cannot Open Website - Server not Running.")
            info = _(
                "You must have the local web server running in order to "
                "view your website in a browser. Click on the 'Run' "
                "button to start the server and then try again."
            )
            self.view.show_message(msg, info)
